#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

r"""SEVIRI Level 1.5 HRIT format reader.

Introduction
------------

The ``seviri_l1b_hrit`` reader reads and calibrates MSG-SEVIRI L1.5 image data in HRIT format. The format is explained
in the `MSG Level 1.5 Image Data Format Description`_. The files are usually named as
follows:

.. code-block:: none

    H-000-MSG4__-MSG4________-_________-PRO______-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000001___-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000002___-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000003___-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000004___-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000005___-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000006___-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000007___-201903011200-__
    H-000-MSG4__-MSG4________-IR_108___-000008___-201903011200-__
    H-000-MSG4__-MSG4________-_________-EPI______-201903011200-__

Each image is decomposed into 24 segments (files) for the high-resolution-visible (HRV) channel and 8 segments for other
visible (VIS) and infrared (IR) channels. Additionally, there is one prologue and one epilogue file for the entire scan
which contain global metadata valid for all channels.

Reader Arguments
----------------
Some arguments can be provided to the reader to change its behaviour. These are
provided through the `Scene` instantiation, eg::

  scn = Scene(filenames=filenames, reader="seviri_l1b_hrit", reader_kwargs={'fill_hrv': False})

To see the full list of arguments that can be provided, look into the documentation
of :class:`HRITMSGFileHandler`.

Compression
-----------

This reader accepts compressed HRIT files, ending in ``C_`` as other HRIT readers, see
:class:`satpy.readers.core.hrit.HRITFileHandler`.

This reader also accepts bzipped file with the extension ``.bz2`` for the prologue,
epilogue, and segment files.

Nominal start/end time
----------------------

.. warning:: attribute access change

``nominal_start_time`` and ``nominal_end_time`` should be accessed using the ``time_parameters`` attribute.

``nominal_start_time`` and ``nominal_end_time`` are also available directly
via ``start_time`` and ``end_time`` respectively.

Here is an exmaple of the content of the start/end time and ``time_parameters`` attibutes

.. code-block:: python

    Start time: 2019-08-29 12:00:00
    End time:   2019-08-29 12:15:00
    time_parameters:
                    {'nominal_start_time': datetime.datetime(2019, 8, 29, 12, 0),
                     'nominal_end_time': datetime.datetime(2019, 8, 29, 12, 15),
                     'observation_start_time': datetime.datetime(2019, 8, 29, 12, 0, 9, 338000),
                     'observation_end_time': datetime.datetime(2019, 8, 29, 12, 15, 9, 203000)
                     }


Example:
--------
Here is an example how to read the data in satpy:

.. code-block:: python

    from satpy import Scene
    import glob

    filenames = glob.glob('data/H-000-MSG4__-MSG4________-*201903011200*')
    scn = Scene(filenames=filenames, reader='seviri_l1b_hrit')
    scn.load(['VIS006', 'IR_108'])
    print(scn['IR_108'])


Output:

.. code-block:: none

    <xarray.DataArray (y: 3712, x: 3712)>
    dask.array<shape=(3712, 3712), dtype=float32, chunksize=(464, 3712)>
    Coordinates:
        acq_time  (y) datetime64[ns] NaT NaT NaT NaT NaT NaT ... NaT NaT NaT NaT NaT
      * x         (x) float64 5.566e+06 5.563e+06 5.56e+06 ... -5.566e+06 -5.569e+06
      * y         (y) float64 -5.566e+06 -5.563e+06 ... 5.566e+06 5.569e+06
    Attributes:
        orbital_parameters:       {'projection_longitude': 0.0, 'projection_latit...
        platform_name:            Meteosat-11
        georef_offset_corrected:  True
        standard_name:            brightness_temperature
        raw_metadata:             {'file_type': 0, 'total_header_length': 6198, '...
        wavelength:               (9.8, 10.8, 11.8)
        units:                    K
        sensor:                   seviri
        platform_name:            Meteosat-11
        start_time:               2019-03-01 12:00:09.716000
        end_time:                 2019-03-01 12:12:42.946000
        area:                     Area ID: some_area_name\\nDescription: On-the-fl...
        name:                     IR_108
        resolution:               3000.403165817
        calibration:              brightness_temperature
        polarization:             None
        level:                    None
        modifiers:                ()
        ancillary_variables:      []

The `filenames` argument can either be a list of strings, see the example above, or a list of
:class:`satpy.readers.core.remote.FSFile` objects. FSFiles can be used in conjunction with `fsspec`_,
e.g. to handle in-memory data:

.. code-block:: python

    import glob

    from fsspec.implementations.memory import MemoryFile, MemoryFileSystem
    from satpy import Scene
    from satpy.readers.core.remote import FSFile

    # In this example, we will make use of `MemoryFile`s in a `MemoryFileSystem`.
    memory_fs = MemoryFileSystem()

    # Usually, the data already resides in memory.
    # For explanatory reasons, we will load the files found with glob in memory,
    #  and load the scene with FSFiles.
    filenames = glob.glob('data/H-000-MSG4__-MSG4________-*201903011200*')
    fs_files = []
    for fn in filenames:
        with open(fn, 'rb') as fh:
            fs_files.append(MemoryFile(
                fs=memory_fs,
                path="{}{}".format(memory_fs.root_marker, fn),
                data=fh.read()
            ))
            fs_files[-1].commit()  # commit the file to the filesystem
    fs_files = [FSFile(open_file) for open_file in filenames]  # wrap MemoryFiles as FSFiles
    # similar to the example above, we pass a list of FSFiles to the `Scene`
    scn = Scene(filenames=fs_files, reader='seviri_l1b_hrit')
    scn.load(['VIS006', 'IR_108'])
    print(scn['IR_108'])


Output:

.. code-block:: none

    <xarray.DataArray (y: 3712, x: 3712)>
    dask.array<shape=(3712, 3712), dtype=float32, chunksize=(464, 3712)>
    Coordinates:
        acq_time  (y) datetime64[ns] NaT NaT NaT NaT NaT NaT ... NaT NaT NaT NaT NaT
      * x         (x) float64 5.566e+06 5.563e+06 5.56e+06 ... -5.566e+06 -5.569e+06
      * y         (y) float64 -5.566e+06 -5.563e+06 ... 5.566e+06 5.569e+06
    Attributes:
        orbital_parameters:       {'projection_longitude': 0.0, 'projection_latit...
        platform_name:            Meteosat-11
        georef_offset_corrected:  True
        standard_name:            brightness_temperature
        raw_metadata:             {'file_type': 0, 'total_header_length': 6198, '...
        wavelength:               (9.8, 10.8, 11.8)
        units:                    K
        sensor:                   seviri
        platform_name:            Meteosat-11
        start_time:               2019-03-01 12:00:09.716000
        end_time:                 2019-03-01 12:12:42.946000
        area:                     Area ID: some_area_name\\nDescription: On-the-fl...
        name:                     IR_108
        resolution:               3000.403165817
        calibration:              brightness_temperature
        polarization:             None
        level:                    None
        modifiers:                ()
        ancillary_variables:      []


References:
    - `EUMETSAT Product Navigator`_
    - `MSG Level 1.5 Image Data Format Description`_
    - `fsspec`_

.. _EUMETSAT Product Navigator:
    https://navigator.eumetsat.int/product/EO:EUM:DAT:MSG:HRSEVIRI
.. _MSG Level 1.5 Image Data Format Description:
    https://user.eumetsat.int/s3/eup-strapi-media/pdf_ten_05105_msg_img_data_e7c8b315e6.pdf
.. _fsspec:
    https://filesystem-spec.readthedocs.io
"""

from __future__ import division

import copy
import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr
from pyresample import geometry

import satpy.readers.core.utils as utils
from satpy._compat import cached_property
from satpy.readers.core._geos_area import get_area_definition, get_area_extent, get_geos_area_naming
from satpy.readers.core.eum import get_service_mode, recarray2dict, time_cds_short
from satpy.readers.core.hrit import (
    HRITFileHandler,
    ancillary_text,
    annotation_header,
    base_hdr_map,
    image_data_function,
)
from satpy.readers.core.seviri import (
    CHANNEL_NAMES,
    HRV_NUM_COLUMNS,
    REPEAT_CYCLE_DURATION,
    SATNUM,
    CalibParams,
    GsicsCoefficients,
    MeirinkCoefficients,
    NominalCoefficients,
    NoValidOrbitParams,
    OrbitPolynomialFinder,
    ScanParams,
    SEVIRICalibrationHandler,
    add_scanline_acq_time,
    create_coef_dict,
    get_cds_time,
    get_satpos,
    mask_bad_quality,
    pad_data_horizontally,
    round_nom_time,
)
from satpy.readers.seviri_l1b_native_hdr import hrit_epilogue, hrit_prologue, impf_configuration
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()
logger = logging.getLogger("hrit_msg")

# MSG implementation:
key_header = np.dtype([("key_number", "u1"),
                       ("seed", ">f8")])

segment_identification = np.dtype([("GP_SC_ID", ">i2"),
                                   ("spectral_channel_id", ">i1"),
                                   ("segment_sequence_number", ">u2"),
                                   ("planned_start_segment_number", ">u2"),
                                   ("planned_end_segment_number", ">u2"),
                                   ("data_field_representation", ">i1")])

image_segment_line_quality = np.dtype([("line_number_in_grid", ">i4"),
                                       ("line_mean_acquisition",
                                        [("days", ">u2"),
                                         ("milliseconds", ">u4")]),
                                       ("line_validity", "u1"),
                                       ("line_radiometric_quality", "u1"),
                                       ("line_geometric_quality", "u1")])

msg_variable_length_headers = {
    image_segment_line_quality: "image_segment_line_quality"}

msg_text_headers = {image_data_function: "image_data_function",
                    annotation_header: "annotation_header",
                    ancillary_text: "ancillary_text"}

msg_hdr_map = base_hdr_map.copy()
msg_hdr_map.update({7: key_header,
                    128: segment_identification,
                    129: image_segment_line_quality
                    })


orbit_coef = np.dtype([("StartTime", time_cds_short),
                       ("EndTime", time_cds_short),
                       ("X", ">f8", (8, )),
                       ("Y", ">f8", (8, )),
                       ("Z", ">f8", (8, )),
                       ("VX", ">f8", (8, )),
                       ("VY", ">f8", (8, )),
                       ("VZ", ">f8", (8, ))])

attitude_coef = np.dtype([("StartTime", time_cds_short),
                          ("EndTime", time_cds_short),
                          ("XofSpinAxis", ">f8", (8, )),
                          ("YofSpinAxis", ">f8", (8, )),
                          ("ZofSpinAxis", ">f8", (8, ))])

cuc_time = np.dtype([("coarse", "u1", (4, )),
                     ("fine", "u1", (3, ))])


class HRITMSGPrologueEpilogueBase(HRITFileHandler):
    """Base reader for prologue and epilogue files."""

    def __init__(self, filename, filename_info, filetype_info, hdr_info):
        """Initialize the file handler for prologue and epilogue files."""
        super().__init__(filename, filename_info, filetype_info, hdr_info)
        self._reduced = None

    def _reduce(self, mda, max_size):
        """Reduce the metadata."""
        if self._reduced is None:
            self._reduced = utils.reduce_mda(mda, max_size=max_size)
        return self._reduced

    def reduce(self, max_size):
        """Reduce the metadata (placeholder)."""
        raise NotImplementedError


class HRITMSGPrologueFileHandler(HRITMSGPrologueEpilogueBase):
    """SEVIRI HRIT prologue reader."""

    def __init__(self, filename, filename_info, filetype_info, calib_mode="nominal",
                 ext_calib_coefs=None, include_raw_metadata=False,
                 mda_max_array_size=None, fill_hrv=None, mask_bad_quality_scan_lines=None):
        """Initialize the reader."""
        super().__init__(filename, filename_info,
                         filetype_info,
                         (msg_hdr_map,
                          msg_variable_length_headers,
                          msg_text_headers))
        self.prologue = {}
        self.read_prologue()

        service = filename_info["service"]
        if service == "":
            self.mda["service"] = "0DEG"
        else:
            self.mda["service"] = service

    def read_prologue(self):
        """Read the prologue metadata."""
        with utils.generic_open(self.filename, mode="rb") as fp_:
            fp_.seek(self.mda["total_header_length"])
            data = np.frombuffer(fp_.read(hrit_prologue.itemsize), dtype=hrit_prologue, count=1)
            self.prologue.update(recarray2dict(data))
            try:
                impf = np.frombuffer(fp_.read(impf_configuration.itemsize), dtype=impf_configuration, count=1)[0]
            except ValueError:
                logger.info("No IMPF configuration field found in prologue.")
            else:
                self.prologue.update(recarray2dict(impf))

    @cached_property
    def satpos(self):
        """Get actual satellite position in geodetic coordinates (WGS-84).

        Evaluate orbit polynomials at the start time of the scan.

        Returns: Longitude [deg east], Latitude [deg north] and Altitude [m]
        """
        a, b = self.get_earth_radii()
        poly_finder = OrbitPolynomialFinder(self.prologue["SatelliteStatus"][
            "Orbit"]["OrbitPolynomial"])
        orbit_polynomial = poly_finder.get_orbit_polynomial(self.observation_start_time)
        return get_satpos(
            orbit_polynomial=orbit_polynomial,
            time=self.observation_start_time,
            semi_major_axis=a,
            semi_minor_axis=b,
        )

    def get_earth_radii(self):
        """Get earth radii from prologue.

        Returns:
            Equatorial radius, polar radius [m]

        """
        earth_model = self.prologue["GeometricProcessing"]["EarthModel"]
        a = earth_model["EquatorialRadius"] * 1000
        b = (earth_model["NorthPolarRadius"] +
             earth_model["SouthPolarRadius"]) / 2.0 * 1000
        return a, b

    def reduce(self, max_size):
        """Reduce the prologue metadata."""
        return self._reduce(self.prologue, max_size=max_size)


class HRITMSGEpilogueFileHandler(HRITMSGPrologueEpilogueBase):
    """SEVIRI HRIT epilogue reader."""

    def __init__(self, filename, filename_info, filetype_info, calib_mode="nominal",
                 ext_calib_coefs=None, include_raw_metadata=False,
                 mda_max_array_size=None, fill_hrv=None, mask_bad_quality_scan_lines=None):
        """Initialize the reader."""
        super(HRITMSGEpilogueFileHandler, self).__init__(filename, filename_info,
                                                         filetype_info,
                                                         (msg_hdr_map,
                                                          msg_variable_length_headers,
                                                          msg_text_headers))
        self.epilogue = {}
        self.read_epilogue()

        service = filename_info["service"]
        if service == "":
            self.mda["service"] = "0DEG"
        else:
            self.mda["service"] = service

    def read_epilogue(self):
        """Read the epilogue metadata."""
        with utils.generic_open(self.filename, mode="rb") as fp_:
            fp_.seek(self.mda["total_header_length"])
            data = np.frombuffer(fp_.read(hrit_epilogue.itemsize), dtype=hrit_epilogue, count=1)
            self.epilogue.update(recarray2dict(data))

    def reduce(self, max_size):
        """Reduce the epilogue metadata."""
        return self._reduce(self.epilogue, max_size=max_size)


class HRITMSGFileHandler(HRITFileHandler):
    """SEVIRI HRIT format reader.

    **Calibration**

    See :mod:`satpy.readers.core.seviri`.


    **Padding of the HRV channel**

    By default, the HRV channel is loaded padded with no-data, returning
    a full-disk dataset. If you want the original, unpadded data,
    just provide the `fill_hrv` as False in the `reader_kwargs`::

        scene = satpy.Scene(filenames,
                            reader='seviri_l1b_hrit',
                            reader_kwargs={'fill_hrv': False})

    **Metadata**

    See :mod:`satpy.readers.core.seviri`.

    """

    def __init__(self, filename, filename_info, filetype_info,
                 prologue, epilogue, calib_mode="nominal",
                 ext_calib_coefs=None, include_raw_metadata=False,
                 mda_max_array_size=100, fill_hrv=True,
                 mask_bad_quality_scan_lines=True):
        """Initialize the reader."""
        super(HRITMSGFileHandler, self).__init__(filename, filename_info,
                                                 filetype_info,
                                                 (msg_hdr_map,
                                                  msg_variable_length_headers,
                                                  msg_text_headers))

        self.prologue_ = prologue
        self.epilogue_ = epilogue
        self.prologue = prologue.prologue
        self.epilogue = epilogue.epilogue
        self._filename_info = filename_info
        self.include_raw_metadata = include_raw_metadata
        self.mda_max_array_size = mda_max_array_size
        self.fill_hrv = fill_hrv
        self.calib_mode = calib_mode
        self.ext_calib_coefs = ext_calib_coefs or {}
        self.mask_bad_quality_scan_lines = mask_bad_quality_scan_lines
        self._get_header()

    def _get_header(self):
        """Read the header info, and fill the metadata dictionary."""
        earth_model = self.prologue["GeometricProcessing"]["EarthModel"]
        self.mda["offset_corrected"] = earth_model["TypeOfEarthModel"] == 2

        # Projection
        a, b = self.prologue_.get_earth_radii()
        self.mda["projection_parameters"]["a"] = a
        self.mda["projection_parameters"]["b"] = b
        ssp = self.prologue["ImageDescription"][
            "ProjectionDescription"]["LongitudeOfSSP"]
        self.mda["projection_parameters"]["SSP_longitude"] = ssp
        self.mda["projection_parameters"]["SSP_latitude"] = 0.0

        # Orbital parameters
        self.mda["orbital_parameters"]["satellite_nominal_longitude"] = self.prologue["SatelliteStatus"][
            "SatelliteDefinition"]["NominalLongitude"]
        self.mda["orbital_parameters"]["satellite_nominal_latitude"] = 0.0
        try:
            actual_lon, actual_lat, actual_alt = self.prologue_.satpos
            self.mda["orbital_parameters"]["satellite_actual_longitude"] = actual_lon
            self.mda["orbital_parameters"]["satellite_actual_latitude"] = actual_lat
            self.mda["orbital_parameters"]["satellite_actual_altitude"] = actual_alt
        except NoValidOrbitParams as err:
            logger.warning(err)

        # Misc
        self.platform_id = self.prologue["SatelliteStatus"][
            "SatelliteDefinition"]["SatelliteId"]
        self.platform_name = "Meteosat-" + SATNUM[self.platform_id]
        self.mda["platform_name"] = self.platform_name
        service = self._filename_info["service"]
        if service == "":
            self.mda["service"] = "0DEG"
        else:
            self.mda["service"] = service
        self.channel_name = CHANNEL_NAMES[self.mda["spectral_channel_id"]]

    @property
    def _repeat_cycle_duration(self):
        """Get repeat cycle duration from epilogue."""
        if self.epilogue["ImageProductionStats"]["ActualScanningSummary"]["ReducedScan"] == 1:
            return 5
        return REPEAT_CYCLE_DURATION

    @property
    def nominal_start_time(self):
        """Get the start time and round it according to scan law."""
        tm = self.prologue["ImageAcquisition"][
            "PlannedAcquisitionTime"]["TrueRepeatCycleStart"]
        return round_nom_time(tm, time_delta=dt.timedelta(minutes=self._repeat_cycle_duration))

    @property
    def nominal_end_time(self):
        """Get the end time and round it according to scan law."""
        tm = self.prologue["ImageAcquisition"][
            "PlannedAcquisitionTime"]["PlannedRepeatCycleEnd"]
        return round_nom_time(tm, time_delta=dt.timedelta(minutes=self._repeat_cycle_duration))

    @property
    def observation_start_time(self):
        """Get the observation start time."""
        return self.epilogue["ImageProductionStats"][
            "ActualScanningSummary"]["ForwardScanStart"]

    @property
    def observation_end_time(self):
        """Get the observation end time."""
        return self.epilogue["ImageProductionStats"][
            "ActualScanningSummary"]["ForwardScanEnd"]

    @property
    def start_time(self):
        """Get general start time for this file."""
        return self.nominal_start_time

    @property
    def end_time(self):
        """Get the general end time for this file."""
        return self.nominal_end_time

    def _get_area_extent(self, pdict):
        """Get the area extent of the file.

        Until December 2017, the data is shifted by 1.5km SSP North and West against the nominal GEOS projection. Since
        December 2017 this offset has been corrected. A flag in the data indicates if the correction has been applied.
        If no correction was applied, adjust the area extent to match the shifted data.

        For more information see Section 3.1.4.2 in the MSG Level 1.5 Image Data Format Description. The correction
        of the area extent is documented in a `developer's memo <https://github.com/pytroll/satpy/wiki/
        SEVIRI-georeferencing-offset-correction>`_.
        """
        aex = get_area_extent(pdict)

        if not self.mda["offset_corrected"]:
            # Geo-referencing offset present. Adjust area extent to match the shifted data. Note that we have to adjust
            # the corners in the *opposite* direction, i.e. S-E. Think of it as if the coastlines were fixed, and you
            # dragged the image to S-E until coastlines and data area aligned correctly.
            #
            # Although the image is flipped upside-down and left-right, the projection coordinates retain their
            # properties, i.e. positive x/y is East/North, respectively.
            xadj = 1500
            yadj = -1500
            aex = (aex[0] + xadj, aex[1] + yadj,
                   aex[2] + xadj, aex[3] + yadj)

        return aex

    def get_area_def(self, dsid):
        """Get the area definition of the band."""
        # Common parameters for both HRV and other channels
        nlines = int(self.mda["number_of_lines"])
        loff = np.float32(self.mda["loff"])
        pdict = dict()
        pdict["cfac"] = np.int32(self.mda["cfac"])
        pdict["lfac"] = np.int32(self.mda["lfac"])
        pdict["coff"] = np.float32(self.mda["coff"])

        pdict["a"] = self.mda["projection_parameters"]["a"]
        pdict["b"] = self.mda["projection_parameters"]["b"]
        pdict["h"] = self.mda["projection_parameters"]["h"]
        pdict["ssp_lon"] = self.mda["projection_parameters"]["SSP_longitude"]

        pdict["nlines"] = nlines
        pdict["ncols"] = int(self.mda["number_of_columns"])
        if (self.prologue["ImageDescription"]["Level15ImageProduction"]
                         ["ImageProcDirection"] == 0):
            pdict["scandir"] = "N2S"
        else:
            pdict["scandir"] = "S2N"

        area_naming_input_dict = {"platform_name": "msg",
                                  "instrument_name": "seviri",
                                  "resolution": int(dsid["resolution"])
                                  }
        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode("seviri", pdict["ssp_lon"])})

        # Compute area definition for non-HRV channels:
        if dsid["name"] != "HRV":
            pdict["loff"] = loff - nlines
            aex = self._get_area_extent(pdict)
            pdict["a_name"] = area_naming["area_id"]
            pdict["a_desc"] = area_naming["description"]
            pdict["p_id"] = ""
            area = get_area_definition(pdict, aex)
            self.area = area
            return self.area

        segment_number = self.mda["segment_sequence_number"]

        current_first_line = ((segment_number -
                               self.mda["planned_start_segment_number"])
                              * pdict["nlines"])

        # Or, if we are processing HRV:
        pdict["a_name"] = area_naming["area_id"]
        pdict["p_id"] = ""
        bounds = self.epilogue["ImageProductionStats"]["ActualL15CoverageHRV"].copy()
        if self.fill_hrv:
            bounds["UpperEastColumnActual"] = 1
            bounds["UpperWestColumnActual"] = HRV_NUM_COLUMNS
            bounds["LowerEastColumnActual"] = 1
            bounds["LowerWestColumnActual"] = HRV_NUM_COLUMNS
            pdict["ncols"] = HRV_NUM_COLUMNS

        upper_south_line = bounds[
            "LowerNorthLineActual"] - current_first_line - 1
        upper_south_line = min(max(upper_south_line, 0), pdict["nlines"])
        lower_coff = (5566 - bounds["LowerEastColumnActual"] + 1)
        upper_coff = (5566 - bounds["UpperEastColumnActual"] + 1)

        # First we look at the lower window
        pdict["nlines"] = upper_south_line
        pdict["loff"] = loff - upper_south_line
        pdict["coff"] = lower_coff
        pdict["a_desc"] = area_naming["description"]
        lower_area_extent = self._get_area_extent(pdict)
        lower_area = get_area_definition(pdict, lower_area_extent)

        # Now the upper window
        pdict["nlines"] = nlines - upper_south_line
        pdict["loff"] = loff - pdict["nlines"] - upper_south_line
        pdict["coff"] = upper_coff
        pdict["a_desc"] = area_naming["description"]
        upper_area_extent = self._get_area_extent(pdict)
        upper_area = get_area_definition(pdict, upper_area_extent)

        area = geometry.StackedAreaDefinition(lower_area, upper_area)

        self.area = area.squeeze()
        return self.area

    def get_dataset(self, key, info):
        """Get the dataset."""
        res = super(HRITMSGFileHandler, self).get_dataset(key, info)
        res = self.calibrate(res, key["calibration"])

        is_calibration = key["calibration"] in ["radiance", "reflectance", "brightness_temperature"]
        if is_calibration and self.mask_bad_quality_scan_lines:
            res = self._mask_bad_quality(res)

        if key["name"] == "HRV" and self.fill_hrv:
            res = self.pad_hrv_data(res)
        self._update_attrs(res, info)
        self._add_scanline_acq_time(res)
        return res

    def pad_hrv_data(self, res):
        """Add empty pixels around the HRV."""
        logger.debug("Padding HRV data to full disk")
        nlines = int(self.mda["number_of_lines"])

        segment_number = self.mda["segment_sequence_number"]

        current_first_line = (segment_number
                              - self.mda["planned_start_segment_number"]) * nlines
        bounds = self.epilogue["ImageProductionStats"]["ActualL15CoverageHRV"]

        upper_south_line = bounds[
          "LowerNorthLineActual"] - current_first_line - 1
        upper_south_line = min(max(upper_south_line, 0), nlines)

        data_list = list()
        if upper_south_line > 0:
            # we have some of the lower window
            data_lower = pad_data_horizontally(res[:upper_south_line, :].data,
                                               (upper_south_line, HRV_NUM_COLUMNS),
                                               bounds["LowerEastColumnActual"],
                                               bounds["LowerWestColumnActual"])
            data_list.append(data_lower)

        if upper_south_line < nlines:
            # we have some of the upper window
            data_upper = pad_data_horizontally(res[upper_south_line:, :].data,
                                               (nlines - upper_south_line, HRV_NUM_COLUMNS),
                                               bounds["UpperEastColumnActual"],
                                               bounds["UpperWestColumnActual"])
            data_list.append(data_upper)
        return xr.DataArray(da.vstack(data_list), dims=("y", "x"), attrs=res.attrs.copy())

    def calibrate(self, data, calibration):
        """Calibrate the data."""
        calib = self._get_calibration_handler()
        res = calib.calibrate(data, calibration)
        return res

    def _get_calibration_handler(self):
        calib_params = CalibParams(
            mode=self.calib_mode.upper(),
            internal_coefs=self._get_calib_coefs(),
            external_coefs=self.ext_calib_coefs,
            radiance_type=self._get_radiance_type()
        )
        scan_params = ScanParams(self.platform_id, self.channel_name,
                                 self.observation_start_time)
        return SEVIRICalibrationHandler(calib_params, scan_params)

    def _mask_bad_quality(self, data):
        """Mask scanlines with bad quality."""
        line_validity = self.mda["image_segment_line_quality"]["line_validity"]
        line_radiometric_quality = self.mda["image_segment_line_quality"]["line_radiometric_quality"]
        line_geometric_quality = self.mda["image_segment_line_quality"]["line_geometric_quality"]
        data = mask_bad_quality(data, line_validity, line_geometric_quality, line_radiometric_quality)
        return data

    def _get_raw_mda(self):
        """Compile raw metadata to be included in the dataset attributes."""
        # Metadata from segment header (excluding items which vary among the different segments)
        raw_mda = copy.deepcopy(self.mda)
        for key in ("image_segment_line_quality", "segment_sequence_number", "annotation_header", "loff"):
            raw_mda.pop(key, None)

        # Metadata from prologue and epilogue (large arrays removed)
        raw_mda.update(self.prologue_.reduce(self.mda_max_array_size))
        raw_mda.update(self.epilogue_.reduce(self.mda_max_array_size))

        return raw_mda

    def _add_scanline_acq_time(self, dataset):
        """Add scanline acquisition time to the given dataset."""
        tline = self.mda["image_segment_line_quality"]["line_mean_acquisition"]
        acq_time = get_cds_time(days=tline["days"], msecs=tline["milliseconds"])
        add_scanline_acq_time(dataset, acq_time)

    def _update_attrs(self, res, info):
        """Update dataset attributes."""
        res.attrs["units"] = info["units"]
        res.attrs["wavelength"] = info["wavelength"]
        res.attrs["standard_name"] = info["standard_name"]
        res.attrs["platform_name"] = self.platform_name
        res.attrs["sensor"] = "seviri"
        res.attrs["nominal_start_time"] = self.nominal_start_time
        res.attrs["nominal_end_time"] = self.nominal_end_time
        res.attrs["time_parameters"] = {
            "nominal_start_time": self.nominal_start_time,
            "nominal_end_time": self.nominal_end_time,
            "observation_start_time": self.observation_start_time,
            "observation_end_time": self.observation_end_time,
            }
        res.attrs["orbital_parameters"] = {
            "projection_longitude": self.mda["projection_parameters"]["SSP_longitude"],
            "projection_latitude": self.mda["projection_parameters"]["SSP_latitude"],
            "projection_altitude": self.mda["projection_parameters"]["h"]}
        res.attrs["orbital_parameters"].update(self.mda["orbital_parameters"])
        res.attrs["georef_offset_corrected"] = self.mda["offset_corrected"]
        if self.include_raw_metadata:
            res.attrs["raw_metadata"] = self._get_raw_mda()

    def _get_calib_coefs(self):
        """Get coefficients for calibration from counts to radiance."""
        band_idx = self._get_band_index()
        coefs_nominal = self.prologue["RadiometricProcessing"][
            "Level15ImageCalibration"]
        coefs_gsics = self.prologue["RadiometricProcessing"]["MPEFCalFeedback"]
        return create_coef_dict(
            nominal_coefs=NominalCoefficients(
                self.channel_name,
                coefs_nominal["CalSlope"][band_idx],
                coefs_nominal["CalOffset"][band_idx]
            ),
            gsics_coefs=GsicsCoefficients(
                self.channel_name,
                coefs_gsics["GSICSCalCoeff"][band_idx],
                coefs_gsics["GSICSOffsetCount"][band_idx]
            ),
            meirink_coefs=MeirinkCoefficients(self.platform_id, self.channel_name, self.observation_start_time)
        )

    def _get_radiance_type(self):
        band_idx = self._get_band_index()
        radiance_types = self.prologue["ImageDescription"][
            "Level15ImageProduction"]["PlannedChanProcessing"]
        return radiance_types[band_idx]

    def _get_band_index(self):
        return self.mda["spectral_channel_id"] - 1


def pad_data(data, final_size, east_bound, west_bound):
    """Pad the data given east and west bounds and the desired size."""
    nlines = final_size[0]
    if west_bound - east_bound != data.shape[1] - 1:
        raise IndexError("East and west bounds do not match data shape")
    padding_east = da.zeros((nlines, east_bound - 1),
                            dtype=data.dtype, chunks=CHUNK_SIZE)
    padding_west = da.zeros((nlines, (final_size[1] - west_bound)),
                            dtype=data.dtype, chunks=CHUNK_SIZE)
    if np.issubdtype(data.dtype, np.floating):
        padding_east = padding_east * np.nan
        padding_west = padding_west * np.nan
    return np.hstack((padding_east, data, padding_west))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for modifiers in modifiers/__init__.py."""
import datetime as dt
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition, StackedAreaDefinition
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy.tests.utils import RANDOM_GEN


def _sunz_area_def():
    """Get fake area for testing sunz generation."""
    area = AreaDefinition("test", "test", "test",
                          {"proj": "merc"}, 2, 2,
                          (-2000, -2000, 2000, 2000))
    return area


def _sunz_bigger_area_def():
    """Get area that is twice the size of 'sunz_area_def'."""
    bigger_area = AreaDefinition("test", "test", "test",
                                 {"proj": "merc"}, 4, 4,
                                 (-2000, -2000, 2000, 2000))
    return bigger_area


def _sunz_stacked_area_def():
    """Get fake stacked area for testing sunz generation."""
    area1 = AreaDefinition("test", "test", "test",
                           {"proj": "merc"}, 2, 1,
                           (-2000, 0, 2000, 2000))
    area2 = AreaDefinition("test", "test", "test",
                           {"proj": "merc"}, 2, 1,
                           (-2000, -2000, 2000, 0))
    return StackedAreaDefinition(area1, area2)


def _shared_sunz_attrs(area_def):
    attrs = {"area": area_def,
             "start_time": dt.datetime(2018, 1, 1, 18),
             "modifiers": tuple(),
             "name": "test_vis"}
    return attrs


def _get_ds1(attrs):
    ds1 = xr.DataArray(da.ones((2, 2), chunks=2, dtype=np.float64),
                       attrs=attrs, dims=("y", "x"),
                       coords={"y": [0, 1], "x": [0, 1]})
    return ds1


@pytest.fixture(scope="session")
def sunz_ds1():
    """Generate fake dataset for sunz tests."""
    attrs = _shared_sunz_attrs(_sunz_area_def())
    return _get_ds1(attrs)


@pytest.fixture(scope="session")
def sunz_ds1_stacked():
    """Generate fake dataset for sunz tests."""
    attrs = _shared_sunz_attrs(_sunz_stacked_area_def())
    return _get_ds1(attrs)


@pytest.fixture(scope="session")
def sunz_ds2():
    """Generate larger fake dataset for sunz tests."""
    attrs = _shared_sunz_attrs(_sunz_bigger_area_def())
    ds2 = xr.DataArray(da.ones((4, 4), chunks=2, dtype=np.float64),
                       attrs=attrs, dims=("y", "x"),
                       coords={"y": [0, 0.5, 1, 1.5], "x": [0, 0.5, 1, 1.5]})
    return ds2


@pytest.fixture(scope="session")
def sunz_sza():
    """Generate fake solar zenith angle data array for testing."""
    sza = xr.DataArray(
        np.rad2deg(np.arccos(da.from_array([[0.0149581333, 0.0146694376], [0.0150812684, 0.0147925727]],
                                           chunks=2))),
        attrs={"area": _sunz_area_def()},
        dims=("y", "x"),
        coords={"y": [0, 1], "x": [0, 1]},
    )
    return sza


class TestSunZenithCorrector:
    """Test case for the zenith corrector."""

    @pytest.mark.parametrize("as_32bit", [False, True])
    def test_basic_default_not_provided(self, sunz_ds1, as_32bit):
        """Test default limits when SZA isn't provided."""
        from satpy.modifiers.geometry import SunZenithCorrector

        if as_32bit:
            sunz_ds1 = sunz_ds1.astype(np.float32)
        comp = SunZenithCorrector(name="sza_test", modifiers=tuple())
        res = comp((sunz_ds1,), test_attr="test")
        np.testing.assert_allclose(res.values, np.array([[22.401667, 22.31777], [22.437503, 22.353533]]),
                                   rtol=1e-6)
        assert "y" in res.coords
        assert "x" in res.coords
        ds1 = sunz_ds1.copy().drop_vars(("y", "x"))
        res = comp((ds1,), test_attr="test")
        res_np = res.compute()
        np.testing.assert_allclose(res_np.values, np.array([[22.401667, 22.31777], [22.437503, 22.353533]]),
                                   rtol=1e-6)
        assert res.dtype == res_np.dtype
        assert "y" not in res.coords
        assert "x" not in res.coords
        if as_32bit:
            assert res.dtype == np.float32

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_basic_lims_not_provided(self, sunz_ds1, dtype):
        """Test custom limits when SZA isn't provided."""
        from satpy.modifiers.geometry import SunZenithCorrector
        comp = SunZenithCorrector(name="sza_test", modifiers=tuple(), correction_limit=90)
        res = comp((sunz_ds1.astype(dtype),), test_attr="test")
        expected = np.array([[66.853262, 68.168939], [66.30742, 67.601493]], dtype=dtype)
        values = res.values
        np.testing.assert_allclose(values, expected, rtol=1e-5)
        assert res.dtype == dtype
        assert values.dtype == dtype

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize("data_arr", [lazy_fixture("sunz_ds1"), lazy_fixture("sunz_ds1_stacked")])
    def test_basic_default_provided(self, data_arr, sunz_sza, dtype):
        """Test default limits when SZA is provided."""
        from satpy.modifiers.geometry import SunZenithCorrector
        comp = SunZenithCorrector(name="sza_test", modifiers=tuple())
        res = comp((data_arr.astype(dtype), sunz_sza.astype(dtype)), test_attr="test")
        expected = np.array([[22.401667, 22.31777], [22.437503, 22.353533]], dtype=dtype)
        values = res.values
        np.testing.assert_allclose(values, expected)
        assert res.dtype == dtype
        assert values.dtype == dtype

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize("data_arr", [lazy_fixture("sunz_ds1"), lazy_fixture("sunz_ds1_stacked")])
    def test_basic_lims_provided(self, data_arr, sunz_sza, dtype):
        """Test custom limits when SZA is provided."""
        from satpy.modifiers.geometry import SunZenithCorrector
        comp = SunZenithCorrector(name="sza_test", modifiers=tuple(), correction_limit=90)
        res = comp((data_arr.astype(dtype), sunz_sza.astype(dtype)), test_attr="test")
        expected = np.array([[66.853262, 68.168939], [66.30742, 67.601493]], dtype=dtype)
        values = res.values
        np.testing.assert_allclose(values, expected, rtol=1e-5)
        assert res.dtype == dtype
        assert values.dtype == dtype

    def test_imcompatible_areas(self, sunz_ds2, sunz_sza):
        """Test sunz correction on incompatible areas."""
        from satpy.composites.core import IncompatibleAreas
        from satpy.modifiers.geometry import SunZenithCorrector
        comp = SunZenithCorrector(name="sza_test", modifiers=tuple(), correction_limit=90)
        with pytest.raises(IncompatibleAreas):
            comp((sunz_ds2, sunz_sza), test_attr="test")


class TestSunZenithReducer:
    """Test case for the sun zenith reducer."""

    @classmethod
    def setup_class(cls):
        """Initialze SunZenithReducer classes that shall be tested."""
        from satpy.modifiers.geometry import SunZenithReducer
        cls.default = SunZenithReducer(name="sza_reduction_test_default", modifiers=tuple())
        cls.custom = SunZenithReducer(name="sza_reduction_test_custom", modifiers=tuple(),
                                      correction_limit=70, max_sza=95, strength=3.0)

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_default_settings(self, sunz_ds1, sunz_sza, dtype):
        """Test default settings with sza data available."""
        res = self.default((sunz_ds1.astype(dtype), sunz_sza.astype(dtype)), test_attr="test")
        expected = np.array([[0.02916261, 0.02839063], [0.02949383, 0.02871911]], dtype=dtype)
        assert res.dtype == dtype
        values = res.values
        assert values.dtype == dtype
        np.testing.assert_allclose(values,
                                   expected,
                                   rtol=2e-5)

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_custom_settings(self, sunz_ds1, sunz_sza, dtype):
        """Test custom settings with sza data available."""
        res = self.custom((sunz_ds1.astype(dtype), sunz_sza.astype(dtype)), test_attr="test")
        expected = np.array([[0.01041319, 0.01030033], [0.01046164, 0.01034834]], dtype=dtype)
        assert res.dtype == dtype
        values = res.values
        assert values.dtype == dtype
        np.testing.assert_allclose(values,
                                   expected,
                                   rtol=1e-5)

    def test_invalid_max_sza(self, sunz_ds1, sunz_sza):
        """Test invalid max_sza with sza data available."""
        from satpy.modifiers.geometry import SunZenithReducer
        with pytest.raises(ValueError, match="`max_sza` must be defined when using the SunZenithReducer."):
            SunZenithReducer(name="sza_reduction_test_invalid", modifiers=tuple(), max_sza=None)


class TestNIRReflectance:
    """Test NIR reflectance compositor."""

    def setup_method(self):
        """Set up the test case for the NIRReflectance compositor."""
        self.area = area = AreaDefinition(
            "test", "", "",
            {"proj": "merc"},
            2,
            2,
            (-2000, -2000, 2000, 2000),
        )
        self.area_hr = AreaDefinition(
            "test", "", "",
            {"proj": "merc"},
            4,
            4,
            (-2000, -2000, 2000, 2000),
        )

        self.start_time = dt.datetime(2020, 1, 1, 12, 0, 0)
        self.metadata = {"platform_name": "Meteosat-11",
                         "sensor": "seviri",
                         "name": "IR_039",
                         "area": area,
                         "start_time": self.start_time}

        self.nir_arr = nir_arr = np.array([[283.15, 285.15], [287.15, 289.15]], dtype=np.float32)
        self.nir = xr.DataArray(da.from_array(nir_arr), dims=["y", "x"])
        self.nir.attrs.update(self.metadata)

        ir_arr = np.array([[273.15, 275.15], [277.15, 279.15]], dtype=np.float32)
        self.ir_ = xr.DataArray(da.from_array(ir_arr), dims=["y", "x"], attrs={"area": area})

        self.sunz_arr = np.array([[1.0, 20.0], [87.0, 89.0]], dtype=np.float32)
        self.sunz = xr.DataArray(da.from_array(self.sunz_arr), dims=["y", "x"],
                                 attrs={"standard_name": "solar_zenith_angle", "area": area})

        co2_arr = np.array([[240.0, 241.0], [242.0, 243.0]], dtype=np.float32)
        self.co2 = xr.DataArray(
            da.from_array(co2_arr),
            dims=("y", "x"),
            attrs={
                "area": self.area,
                "start_time": self.start_time,
                "wavelength": (12.0, 13.0, 14.0),
                "units": "K",
            })


    @pytest.mark.parametrize(
        ("include_sunz", "include_co2", "exp_res"),
        [
            (False, False, np.array([[4.251828, 4.639434], [5.0589, 5.514466]], dtype=np.float32)),
            (True, False, np.array([[3.8915825, 4.5359993], [np.nan, np.nan]], dtype=np.float32)),
            (False, True, np.array([[5.0192585, 5.5059953], [6.0353055, 6.6126623]], dtype=np.float32)),
        ]
    )
    def test_basic_call(self, tmp_path, include_sunz, include_co2, exp_res):
        """Test NIR reflectance compositor with various optional inputs."""
        from pyspectral.testing import mock_tb_conversion

        from satpy.modifiers.spectral import NIRReflectance

        opt_datasets = []
        if include_sunz:
            opt_datasets.append(self.sunz)
        if include_co2:
            opt_datasets.append(self.co2)
        comp = NIRReflectance(name="test")
        info = {"modifiers": None}
        with mock_tb_conversion(tb2rad_dir=tmp_path, central_wavelengths={"IR_039": 3.9}):
            res = comp([self.nir, self.ir_], optional_datasets=opt_datasets, **info)
        res_da = res.data
        res_np = res.data.compute()
        assert res_np.dtype == res_da.dtype
        assert res_np.dtype == self.nir.dtype

        assert comp.sun_zenith_threshold == 85.0
        assert comp.masking_limit == 88.0
        assert self.metadata.items() <= res.attrs.items()
        assert res.attrs["units"] == "%"
        assert res.attrs["sun_zenith_threshold"] == 85.0
        assert res.attrs["sun_zenith_masking_limit"] == 88.0
        np.testing.assert_allclose(res_np, exp_res, atol=2e-6)

    @pytest.mark.parametrize(
        "comp_kwargs",
        [
            {"sunz_threshold": 84.0},
            {"masking_limit": None},
        ]
    )
    def test_provide_sunz_threshold_and_masking_limit(self, tmp_path, comp_kwargs):
        """Test NIR reflectance compositor provided sunz and a sunz threshold."""
        from pyspectral.near_infrared_reflectance import Calculator
        from pyspectral.testing import mock_tb_conversion

        from satpy.modifiers.spectral import NIRReflectance

        comp = NIRReflectance(name="test", **comp_kwargs)
        exp_call_kwargs = {
            "sunz_threshold": comp_kwargs.get("sunz_threshold", NIRReflectance.TERMINATOR_LIMIT),
            "masking_limit": comp_kwargs.get("masking_limit", NIRReflectance.MASKING_LIMIT),
        }
        info = {"modifiers": None}

        with mock_tb_conversion(tb2rad_dir=tmp_path, central_wavelengths={"IR_039": 3.9}), \
                mock.patch("satpy.modifiers.spectral.Calculator", wraps=Calculator) as calculator:
            res = comp([self.nir, self.ir_], optional_datasets=[self.sunz], **info)

        assert res.attrs["sun_zenith_threshold"] == exp_call_kwargs["sunz_threshold"]
        assert res.attrs["sun_zenith_masking_limit"] == exp_call_kwargs["masking_limit"]
        calculator.assert_called_with("Meteosat-11", "seviri", "IR_039", **exp_call_kwargs)

    def test_nir_multiple_resolutions(self):
        """Check that multiple resolutions in the optional datasets produce an IncompatibleArea."""
        from satpy.composites.core import IncompatibleAreas
        from satpy.modifiers.spectral import NIRReflectance

        # make sunz that is twice as many pixels
        sunz_arr = np.array([
            [1.0, 2.0, 3.0, 4.0],
            [3.0, 4.0, 5.0, 6.0],
        ], dtype=np.float32)
        sunz = xr.DataArray(da.from_array(sunz_arr), dims=["y", "x"])
        sunz.attrs["standard_name"] = "solar_zenith_angle"
        sunz.attrs["area"] = self.area_hr

        comp = NIRReflectance(name="test")
        info = {"modifiers": None}
        with pytest.raises(IncompatibleAreas):
            comp([self.nir, self.ir_], optional_datasets=[sunz], **info)


class TestNIREmissivePartFromReflectance(unittest.TestCase):
    """Test the NIR Emissive part from reflectance compositor."""

    @mock.patch("satpy.modifiers.spectral.sun_zenith_angle")
    @mock.patch("satpy.modifiers.NIRReflectance.apply_modifier_info")
    @mock.patch("satpy.modifiers.spectral.Calculator")
    def test_compositor(self, calculator, apply_modifier_info, sza):
        """Test the NIR emissive part from reflectance compositor."""
        from satpy.modifiers.spectral import NIRReflectance

        refl_arr = RANDOM_GEN.random((2, 2))
        refl = da.from_array(refl_arr)

        refl_from_tbs = mock.MagicMock()
        refl_from_tbs.return_value = refl
        calculator.return_value = mock.MagicMock(reflectance_from_tbs=refl_from_tbs)

        emissive_arr = RANDOM_GEN.random((2, 2))
        emissive = da.from_array(emissive_arr)
        emissive_part = mock.MagicMock()
        emissive_part.return_value = emissive
        calculator.return_value = mock.MagicMock(emissive_part_3x=emissive_part)

        from satpy.modifiers.spectral import NIREmissivePartFromReflectance

        comp = NIREmissivePartFromReflectance(name="test", sunz_threshold=86.0)
        info = {"modifiers": None}

        platform = "NOAA-20"
        sensor = "viirs"
        chan_name = "M12"

        get_lonlats = mock.MagicMock()
        lons, lats = 1, 2
        get_lonlats.return_value = (lons, lats)
        area = mock.MagicMock(get_lonlats=get_lonlats)

        nir_arr = RANDOM_GEN.random((2, 2))
        nir = xr.DataArray(da.from_array(nir_arr), dims=["y", "x"])
        nir.attrs["platform_name"] = platform
        nir.attrs["sensor"] = sensor
        nir.attrs["name"] = chan_name
        nir.attrs["area"] = area
        ir_arr = RANDOM_GEN.random((2, 2))
        ir_ = xr.DataArray(da.from_array(ir_arr), dims=["y", "x"])
        ir_.attrs["area"] = area

        sunz_arr = 100 * RANDOM_GEN.random((2, 2))
        sunz = xr.DataArray(da.from_array(sunz_arr), dims=["y", "x"])
        sunz.attrs["standard_name"] = "solar_zenith_angle"
        sunz.attrs["area"] = area
        sunz2 = da.from_array(sunz_arr)
        sza.return_value = sunz2

        res = comp([nir, ir_], optional_datasets=[sunz], **info)
        assert res.attrs["sun_zenith_threshold"] == 86.0
        assert res.attrs["units"] == "K"
        assert res.attrs["platform_name"] == platform
        assert res.attrs["sensor"] == sensor
        assert res.attrs["name"] == chan_name
        calculator.assert_called_with("NOAA-20", "viirs", "M12", sunz_threshold=86.0,
                                      masking_limit=NIRReflectance.MASKING_LIMIT)


class TestPSPRayleighReflectance:
    """Test the pyspectral-based Rayleigh correction modifier."""

    def _make_data_area(self):
        """Create test area definition and data."""
        rows = 3
        cols = 5
        area = AreaDefinition(
            "some_area_name", "On-the-fly area", "geosabii",
            {"a": "6378137.0", "b": "6356752.31414", "h": "35786023.0", "lon_0": "-89.5", "proj": "geos", "sweep": "x",
             "units": "m"},
            cols, rows,
            (-5434894.954752679, -5434894.964451744, 5434894.964451744, 5434894.954752679))

        data = np.zeros((rows, cols)) + 25
        data[1, :] += 25
        data[2, :] += 50
        data = da.from_array(data, chunks=2)
        return area, data

    def _create_test_data(self, name, wavelength, resolution):
        area, dnb = self._make_data_area()
        input_band = xr.DataArray(dnb,
                                  dims=("y", "x"),
                                  attrs={
                                      "platform_name": "Himawari-8",
                                      "calibration": "reflectance", "units": "%", "wavelength": wavelength,
                                      "name": name, "resolution": resolution, "sensor": "ahi",
                                      "start_time": "2017-09-20 17:30:40.800000",
                                      "end_time": "2017-09-20 17:41:17.500000",
                                      "area": area, "ancillary_variables": [],
                                      "orbital_parameters": {
                                          "satellite_nominal_longitude": -89.5,
                                          "satellite_nominal_latitude": 0.0,
                                          "satellite_nominal_altitude": 35786023.4375,
                                      },
                                  })

        red_band = xr.DataArray(dnb,
                                dims=("y", "x"),
                                attrs={
                                    "platform_name": "Himawari-8",
                                    "calibration": "reflectance", "units": "%", "wavelength": (0.62, 0.64, 0.66),
                                    "name": "B03", "resolution": 500, "sensor": "ahi",
                                    "start_time": "2017-09-20 17:30:40.800000",
                                    "end_time": "2017-09-20 17:41:17.500000",
                                    "area": area, "ancillary_variables": [],
                                    "orbital_parameters": {
                                        "satellite_nominal_longitude": -89.5,
                                        "satellite_nominal_latitude": 0.0,
                                        "satellite_nominal_altitude": 35786023.4375,
                                    },
                                })
        fake_angle_data = da.ones_like(dnb, dtype=np.float32) * 90.0
        angle1 = xr.DataArray(fake_angle_data,
                              dims=("y", "x"),
                              attrs={
                                  "platform_name": "Himawari-8",
                                  "calibration": "reflectance", "units": "%", "wavelength": wavelength,
                                  "name": "satellite_azimuth_angle", "resolution": resolution, "sensor": "ahi",
                                  "start_time": "2017-09-20 17:30:40.800000",
                                  "end_time": "2017-09-20 17:41:17.500000",
                                  "area": area, "ancillary_variables": [],
                              })
        return input_band, red_band, angle1, angle1, angle1, angle1

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize(
        ("name", "wavelength", "resolution", "aerosol_type", "reduce_lim_low", "reduce_lim_high", "reduce_strength"),
        [
            ("B01", (0.45, 0.47, 0.49), 1000, "rayleigh_only", 70, 95, 1),
            ("B02", (0.49, 0.51, 0.53), 1000, "rayleigh_only", 70, 95, 1),
            ("B03", (0.62, 0.64, 0.66), 500, "rayleigh_only", 70, 95, 1),
            ("B01", (0.45, 0.47, 0.49), 1000, "rayleigh_only", -95, -70, -1),
        ]
    )
    def test_rayleigh_corrector(
            self, tmp_path, name, wavelength, resolution, aerosol_type,
            reduce_lim_low, reduce_lim_high, reduce_strength, dtype):
        """Test PSPRayleighReflectance with fake data."""
        from pyspectral.testing import mock_rayleigh

        from satpy.modifiers.atmosphere import PSPRayleighReflectance

        ray_cor = PSPRayleighReflectance(name=name, atmosphere="us-standard", aerosol_types=aerosol_type,
                                         reduce_lim_low=reduce_lim_low, reduce_lim_high=reduce_lim_high,
                                         reduce_strength=reduce_strength)
        assert ray_cor.attrs["name"] == name
        assert ray_cor.attrs["atmosphere"] == "us-standard"
        assert ray_cor.attrs["aerosol_types"] == aerosol_type
        assert ray_cor.attrs["reduce_lim_low"] == reduce_lim_low
        assert ray_cor.attrs["reduce_lim_high"] == reduce_lim_high
        assert ray_cor.attrs["reduce_strength"] == reduce_strength

        input_band, red_band, *_ = self._create_test_data(name, wavelength, resolution)
        with mock_rayleigh(rayleigh_dir=tmp_path):
            res = ray_cor([input_band.astype(dtype), red_band.astype(dtype)])

        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.dtype == dtype
        data = res.values
        assert data.shape == (3, 5)
        assert data.dtype == dtype

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize("as_optionals", [False, True])
    def test_rayleigh_with_angles(self, tmp_path, as_optionals, dtype):
        """Test PSPRayleighReflectance with angles provided."""
        from pyspectral.testing import mock_rayleigh

        from satpy.modifiers.atmosphere import PSPRayleighReflectance

        aerosol_type = "rayleigh_only"
        ray_cor = PSPRayleighReflectance(name="B01", atmosphere="us-standard", aerosol_types=aerosol_type)
        prereqs, opt_prereqs = self._get_angles_prereqs_and_opts(as_optionals, dtype)
        with mock.patch("satpy.modifiers.atmosphere.get_angles") as get_angles, mock_rayleigh(rayleigh_dir=tmp_path):
            res = ray_cor(prereqs, opt_prereqs)
        get_angles.assert_not_called()

        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.dtype == dtype
        data = res.values
        assert data.shape == (3, 5)
        assert data.dtype == dtype

    def _get_angles_prereqs_and_opts(self, as_optionals, dtype):
        wavelength = (0.45, 0.47, 0.49)
        resolution = 1000
        input_band, red_band, *angles = self._create_test_data("B01", wavelength, resolution)
        prereqs = [input_band.astype(dtype), red_band.astype(dtype)]
        opt_prereqs = []
        angles = [a.astype(dtype) for a in angles]
        if as_optionals:
            opt_prereqs = angles
        else:
            prereqs += angles
        return prereqs, opt_prereqs


class TestPSPAtmosphericalCorrection(unittest.TestCase):
    """Test the pyspectral-based atmospheric correction modifier."""

    def test_call(self):
        """Test atmospherical correction."""
        from pyresample.geometry import SwathDefinition

        from satpy.modifiers import PSPAtmosphericalCorrection

        # Patch methods
        lons = np.zeros((5, 5))
        lons[1, 1] = np.inf
        lons = da.from_array(lons, chunks=5)
        lats = np.zeros((5, 5))
        lats[1, 1] = np.inf
        lats = da.from_array(lats, chunks=5)
        area = SwathDefinition(lons, lats)
        stime = dt.datetime(2020, 1, 1, 12, 0, 0)
        orb_params = {
            "satellite_actual_altitude": 12345678,
            "nadir_longitude": 0.0,
            "nadir_latitude": 0.0,
        }
        band = xr.DataArray(da.zeros((5, 5)),
                            attrs={"area": area,
                                   "start_time": stime,
                                   "name": "name",
                                   "platform_name": "platform",
                                   "sensor": "sensor",
                                   "orbital_parameters": orb_params},
                            dims=("y", "x"))

        # Perform atmospherical correction
        psp = PSPAtmosphericalCorrection(name="dummy")
        res = psp(projectables=[band])
        res.compute()



#!/usr/bin/env python
# Copyright (c) 2018-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittests for NWC SAF reader."""

import h5netcdf
import numpy as np
import pytest
import xarray as xr

from satpy.readers.nwcsaf_nc import NcNWCSAF, read_nwcsaf_time
from satpy.tests.utils import RANDOM_GEN

PROJ_KM = {"gdal_projection": "+proj=geos +a=6378.137000 +b=6356.752300 +lon_0=0.000000 +h=35785.863000",
           "gdal_xgeo_up_left": -5569500.0,
           "gdal_ygeo_up_left": 5437500.0,
           "gdal_xgeo_low_right": 5566500.0,
           "gdal_ygeo_low_right": 2653500.0}

NOMINAL_ALTITUDE = 35785863.0

PROJ = {"gdal_projection": f"+proj=geos +a=6378137.000 +b=6356752.300 +lon_0=0.000000 +h={NOMINAL_ALTITUDE:.3f}",
        "gdal_xgeo_up_left": -5569500.0,
        "gdal_ygeo_up_left": 5437500.0,
        "gdal_xgeo_low_right": 5566500.0,
        "gdal_ygeo_low_right": 2653500.0}


dimensions_v2021 = {
    "nx": 1530,
    "ny": 928,
    "pal_colors_250": 250,
    "pal_rgb": 3,
}

dimensions_v2025 = dimensions_v2021.copy()
dimensions_v2025["time"] = 1

NOMINAL_LONGITUDE = 0.0
NOMINAL_TIME = "2023-01-18T10:30:00Z"
START_TIME = "2023-01-18T10:39:17Z"
END_TIME = "2023-01-18T10:42:22Z"
START_TIME_PPS = "20230118T103917000Z"
END_TIME_PPS = "20230118T104222000Z"

global_attrs = {
    "source": "NWC/GEO version v2021.1",
    "satellite_identifier": "MSG4",
    "sub-satellite_longitude": NOMINAL_LONGITUDE,
    "time_coverage_start": START_TIME,
    "time_coverage_end": END_TIME,
}

global_attrs.update(PROJ)

global_attrs_geo = global_attrs.copy()
global_attrs_geo["nominal_product_time"] = NOMINAL_TIME

global_attrs_geo_v2025 = {
    "source": "NWC/GEO version v2025.1",
    "satellite_identifier": "MTI1",
    "sub-satellite_longitude": NOMINAL_LONGITUDE,
    "time_coverage_start": "2025-09-23T13:07:53Z",
    "time_coverage_end": "2025-09-23T13:09:29Z",
}

CTTH_PALETTE_MEANINGS = ("0 500 1000 1500")

COT_PALETTE_MEANINGS = ("0 2 5 8 10 13 16 19 23 26 29 33 36 40 43 47 51 55 59 63 68 72 77 81 86 91 96"
                        " 101 107 112 118 123 129 135 142 148 154 161 168 175 182 190 198 205 213 222"
                        " 230 239 248 257 266 276 286 296 307 317 328 340 351 363 375 388 401 414 428"
                        " 442 456 470 485 501 517 533 550 567 584 602 621 640 660 680 700 721 743 765"
                        " 788 811 835 860 885 911 938 965 993 1022 1052 1082 1113 1145 1178 1212 1246"
                        " 1282 1318 1355 1394 1433 1474 1515 1558 1601 1646 1692 1739 1788 1837 1889 "
                        "1941 1995 2050 2107 2165 2224 2286 2348 2413 2479 2547 2617 2688 2762 2837 "
                        "2915 2994 3076 3159 3245 3333 3424 3517 3612 3710 3810 3913 4019 4127 4239 "
                        "4353 4470 4591 4714 4841 4971 5105 5242 5383 5527 5676 5828 5984 6144 6309 "
                        "6478 6651 6829 7011 7199 7391 7588 7791 7999 8212 8431 8656 8886 9123 9366 "
                        "9615 9871 10134 10404 10680 10964 11256 11555 11862 12177 12501 12833 13173 "
                        "13523 13882 14250 14628 15016 15414 15823 16243 16673 17115 17569 18034 18512"
                        " 19002 19505 20022 20552 21096 21654 22227 22816 23419 24039 24675 25327 "
                        "25997 26685 27390 28114 28858 29621 30404 31207 32032 32878 33747 34639 35554"
                        " 36493 37457 38446 39462 40504 41574 42672 43798 44955 46142 47360 48610 "
                        "49893 51210 52562 53949 55373 56834 58334 59873 61453 63075 64739")

COT_SCALE = 0.01
COT_OFFSET = 0.0

CRE_ARRAY = RANDOM_GEN.integers(0, 65535, size=(928, 1530), dtype=np.uint16)
COT_ARRAY = RANDOM_GEN.integers(0, 65535, size=(928, 1530), dtype=np.uint16)
PAL_ARRAY = RANDOM_GEN.integers(0, 255, size=(250, 3), dtype=np.uint8)


@pytest.fixture(scope="session")
def nwcsaf_geo_ct_filename(tmp_path_factory):
    """Create a CT file and return the filename."""
    return create_nwcsaf_geo_ct_file(tmp_path_factory.mktemp("data"))


def create_nwcsaf_geo_ct_file(directory, attrs=global_attrs_geo):
    """Create a CT file."""
    filename = directory / "S_NWC_CT_MSG4_MSG-N-VISIR_20230118T103000Z_PLAX.nc"
    _create_nwcsaf_cf_file(filename, attrs, "v2021")
    return filename


def _create_nwcsaf_cf_file(filename, attrs, version):
    with h5netcdf.File(filename, mode="w") as nc_file:
        nc_file.attrs.update(attrs)
        var_name = "ct"
        if version == "v2025":
            dim_names = ("time", "ny", "nx")
            chunks = (1, 256, 256)
            dimensions = dimensions_v2025
            shape = (1, 928, 1530)
        else:
            dim_names = ("ny", "nx")
            chunks = (256, 256)
            dimensions = dimensions_v2021
            shape = (928, 1530)
        nc_file.dimensions = dimensions
        var = nc_file.create_variable(var_name, dim_names, np.uint8,
                                      chunks=chunks)
        var[:] = RANDOM_GEN.integers(0, 255, size=shape, dtype=np.uint8)


@pytest.fixture
def nwcsaf_geo_ct_filehandler(nwcsaf_geo_ct_filename):
    """Create a CT filehandler."""
    return NcNWCSAF(nwcsaf_geo_ct_filename, {}, {})


@pytest.fixture(scope="session")
def nwcsaf_geo_v2025_ct_filename(tmp_path_factory):
    """Create a CT file in the v2025 format and return the filename."""
    return create_nwcsaf_geo_v2025_ct_file(tmp_path_factory.mktemp("data"))


def create_nwcsaf_geo_v2025_ct_file(directory, attrs=global_attrs_geo_v2025):
    """Create a CT file in v2025 format."""
    filename = directory / "S_NWC_CT_MTI1_MSG-N-NR_20250923T130000Z.nc"
    _create_nwcsaf_cf_file(filename, attrs, "v2025")
    return filename


@pytest.fixture
def nwcsaf_geo_v2025_ct_filehandler(nwcsaf_geo_v2025_ct_filename):
    """Create a CT filehandler."""
    return NcNWCSAF(nwcsaf_geo_v2025_ct_filename, {}, {})


@pytest.fixture(scope="session")
def nwcsaf_pps_cmic_filename(tmp_path_factory):
    """Create a CMIC file."""
    attrs = global_attrs.copy()
    attrs.update(PROJ_KM)
    attrs["time_coverage_start"] = START_TIME_PPS
    attrs["time_coverage_end"] = END_TIME_PPS
    filename = create_cmic_file(tmp_path_factory.mktemp("data"), filetype="cmic", attrs=attrs)
    return filename


@pytest.fixture(scope="session")
def nwcsaf_pps_ctth_filename(tmp_path_factory):
    """Create a CTTH file."""
    attrs = global_attrs.copy()
    attrs.update(PROJ_KM)
    attrs["time_coverage_start"] = START_TIME_PPS
    attrs["time_coverage_end"] = END_TIME_PPS
    filename = create_ctth_file(tmp_path_factory.mktemp("data"), attrs=attrs)
    return filename


def create_cmic_file(path, filetype, attrs=global_attrs):
    """Create a cmic file."""
    filename = path / f"S_NWC_{filetype.upper()}_npp_00000_20230118T1427508Z_20230118T1429150Z.nc"
    with h5netcdf.File(filename, mode="w") as nc_file:
        nc_file.dimensions = dimensions_v2021
        nc_file.attrs.update(attrs)
        create_cot_variable(nc_file, f"{filetype}_cot")
        create_cot_pal_variable(nc_file, f"{filetype}_cot_pal")
        create_cre_variables(nc_file, f"{filetype}_cre")
    return filename


def create_ctth_file(path, attrs=global_attrs):
    """Create a cmic file."""
    filename = path / "S_NWC_CTTH_npp_00000_20230118T1427508Z_20230118T1429150Z.nc"
    with h5netcdf.File(filename, mode="w") as nc_file:
        nc_file.dimensions = dimensions_v2021
        nc_file.attrs.update(attrs)
        create_ctth_variables(nc_file, "ctth_alti")
        create_ctth_alti_pal_variable_with_fill_value_color(nc_file, "ctth_alti_pal")
    return filename


@pytest.fixture
def nwcsaf_pps_cmic_filehandler(nwcsaf_pps_cmic_filename):
    """Create a CMIC filehandler."""
    return NcNWCSAF(nwcsaf_pps_cmic_filename, {}, {"file_key_prefix": "cmic_"})


@pytest.fixture
def nwcsaf_pps_ctth_filehandler(nwcsaf_pps_ctth_filename):
    """Create a CMIC filehandler."""
    return NcNWCSAF(nwcsaf_pps_ctth_filename, {}, {})


@pytest.fixture(scope="session")
def nwcsaf_pps_cpp_filename(tmp_path_factory):
    """Create a CPP file."""
    filename = create_cmic_file(tmp_path_factory.mktemp("data"), filetype="cpp")

    return filename


def create_cre_variables(nc_file, var_name):
    """Create a CRE variable."""
    var = nc_file.create_variable(var_name, ("ny", "nx"), np.uint16, chunks=(256, 256))
    var[:] = CRE_ARRAY


def create_ctth_variables(nc_file, var_name):
    """Create a CRE variable."""
    var = nc_file.create_variable(var_name, ("ny", "nx"), np.uint16, chunks=(256, 256))
    var[:] = CRE_ARRAY
    var.attrs["scale_factor"] = COT_SCALE
    var.attrs["add_offset"] = COT_OFFSET
    var.attrs["_FillValue"] = 65535


def create_cot_pal_variable(nc_file, var_name):
    """Create a palette variable."""
    var = nc_file.create_variable(var_name, ("pal_colors_250", "pal_rgb"), np.uint8)
    var[:] = PAL_ARRAY
    var.attrs["palette_meanings"] = COT_PALETTE_MEANINGS


def create_cot_variable(nc_file, var_name):
    """Create a COT variable."""
    var = nc_file.create_variable(var_name, ("ny", "nx"), np.uint16, chunks=(256, 256))
    var[:] = COT_ARRAY
    var.attrs["scale_factor"] = COT_SCALE
    var.attrs["add_offset"] = COT_OFFSET
    var.attrs["_FillValue"] = 65535


def create_ctth_alti_pal_variable_with_fill_value_color(nc_file, var_name):
    """Create a palette variable."""
    var = nc_file.create_variable(var_name, ("pal_colors_250", "pal_rgb"), np.uint8)
    var[:] = PAL_ARRAY
    var.attrs["palette_meanings"] = CTTH_PALETTE_MEANINGS
    var.attrs["fill_value_color"] = [0, 0, 0]
    var.attrs["scale_factor"] = COT_SCALE
    var.attrs["add_offset"] = COT_OFFSET
    var.attrs["_FillValue"] = 65535


@pytest.fixture
def nwcsaf_pps_cpp_filehandler(nwcsaf_pps_cpp_filename):
    """Create a CPP filehandler."""
    return NcNWCSAF(nwcsaf_pps_cpp_filename, {}, {"file_key_prefix": "cpp_"})


@pytest.fixture(scope="session")
def nwcsaf_old_geo_ct_filename(tmp_path_factory):
    """Create a CT file and return the filename."""
    attrs = global_attrs_geo.copy()
    attrs.update(PROJ_KM)
    attrs["time_coverage_start"] = np.array(["2023-01-18T10:39:17Z"], dtype="S20")
    return create_nwcsaf_geo_ct_file(tmp_path_factory.mktemp("data-old"), attrs=attrs)


@pytest.fixture
def nwcsaf_old_geo_ct_filehandler(nwcsaf_old_geo_ct_filename):
    """Create a CT filehandler."""
    return NcNWCSAF(nwcsaf_old_geo_ct_filename, {}, {})


class TestNcNWCSAFGeo:
    """Test the NcNWCSAF reader for Geo products."""

    @pytest.mark.parametrize(("platform", "instrument"), [("Metop-B", "avhrr-3"),
                                                          ("NOAA-20", "viirs"),
                                                          ("Himawari-8", "ahi"),
                                                          ("GOES-17", "abi"),
                                                          ("Meteosat-11", "seviri")])
    def test_sensor_name_platform(self, nwcsaf_geo_ct_filehandler, platform, instrument):
        """Test that the correct sensor name is being set."""
        nwcsaf_geo_ct_filehandler.set_platform_and_sensor(platform_name=platform)
        assert nwcsaf_geo_ct_filehandler.sensor == set([instrument])
        assert nwcsaf_geo_ct_filehandler.sensor_names == set([instrument])

    @pytest.mark.parametrize(("platform", "instrument"), [("GOES16", "abi"),
                                                          ("MSG4", "seviri")])
    def test_sensor_name_sat_id(self, nwcsaf_geo_ct_filehandler, platform, instrument):
        """Test that the correct sensor name is being set."""
        nwcsaf_geo_ct_filehandler.set_platform_and_sensor(sat_id=platform)
        assert nwcsaf_geo_ct_filehandler.sensor == set([instrument])
        assert nwcsaf_geo_ct_filehandler.sensor_names == set([instrument])

    def test_get_area_def(self, nwcsaf_geo_ct_filehandler):
        """Test that get_area_def() returns proper area."""
        dsid = {"name": "ct"}
        _check_filehandler_area_def(nwcsaf_geo_ct_filehandler, dsid)

    def test_get_area_def_km(self, nwcsaf_old_geo_ct_filehandler):
        """Test that get_area_def() returns proper area when the projection is in km."""
        dsid = {"name": "ct"}
        _check_filehandler_area_def(nwcsaf_old_geo_ct_filehandler, dsid)

    def test_scale_dataset_attr_removal(self, nwcsaf_geo_ct_filehandler):
        """Test the scaling of the dataset and removal of obsolete attributes."""
        import numpy as np
        import xarray as xr

        attrs = {"scale_factor": np.array(10),
                 "add_offset": np.array(20)}
        var = xr.DataArray([1, 2, 3], attrs=attrs)

        var = nwcsaf_geo_ct_filehandler.scale_dataset(var, "dummy")
        np.testing.assert_allclose(var, [30, 40, 50])
        assert "scale_factor" not in var.attrs
        assert "add_offset" not in var.attrs

    @pytest.mark.parametrize(("attrs", "expected"), [({"scale_factor": np.array(1.5),
                                                       "add_offset": np.array(2.5),
                                                       "_FillValue": 1},
                                                      [np.nan, 5.5, 7]),
                                                     ({"scale_factor": np.array(1.5),
                                                       "add_offset": np.array(2.5),
                                                       "valid_min": 1.1},
                                                      [np.nan, 5.5, 7]),
                                                     ({"scale_factor": np.array(1.5),
                                                       "add_offset": np.array(2.5),
                                                       "valid_max": 2.1},
                                                      [4, 5.5, np.nan]),
                                                     ({"scale_factor": np.array(1.5),
                                                       "add_offset": np.array(2.5),
                                                       "valid_range": (1.1, 2.1)},
                                                      [np.nan, 5.5, np.nan])])
    def test_scale_dataset_floating(self, nwcsaf_geo_ct_filehandler, attrs, expected):
        """Test the scaling of the dataset with floating point values."""
        var = xr.DataArray([1, 2, 3], attrs=attrs)
        var = nwcsaf_geo_ct_filehandler.scale_dataset(var, "dummy")
        np.testing.assert_allclose(var, expected)
        assert "scale_factor" not in var.attrs
        assert "add_offset" not in var.attrs

    def test_scale_dataset_floating_nwcsaf_geo_ctth(self, nwcsaf_geo_ct_filehandler):
        """Test the scaling of the dataset with floating point values for CTTH NWCSAF/Geo v2016/v2018."""
        attrs = {"scale_factor": np.array(1.),
                 "add_offset": np.array(-2000.),
                 "valid_range": (0., 27000.)}
        var = xr.DataArray([1, 2, 3], attrs=attrs)
        var = nwcsaf_geo_ct_filehandler.scale_dataset(var, "dummy")
        np.testing.assert_allclose(var, [-1999., -1998., -1997.])
        assert "scale_factor" not in var.attrs
        assert "add_offset" not in var.attrs
        np.testing.assert_equal(var.attrs["valid_range"], (-2000., 25000.))

    def test_scale_dataset_uint8_noop(self, nwcsaf_geo_ct_filehandler):
        """Test that uint8 is not accidentally casted when no scaling is done."""
        attrs = {}
        var = xr.DataArray(np.array([1, 2, 3], dtype=np.uint8), attrs=attrs)
        var = nwcsaf_geo_ct_filehandler.scale_dataset(var, "dummy")
        np.testing.assert_equal(var, np.array([1, 2, 3], dtype=np.uint8))
        assert var.dtype == np.uint8

    def test_orbital_parameters_are_correct(self, nwcsaf_geo_ct_filehandler):
        """Test that orbital parameters are present in the dataset attributes."""
        dsid = {"name": "ct"}
        var = nwcsaf_geo_ct_filehandler.get_dataset(dsid, {})
        assert "orbital_parameters" in var.attrs
        for param in var.attrs["orbital_parameters"]:
            assert isinstance(var.attrs["orbital_parameters"][param], (float, int))

        assert var.attrs["orbital_parameters"]["satellite_nominal_altitude"] == NOMINAL_ALTITUDE
        assert var.attrs["orbital_parameters"]["satellite_nominal_longitude"] == NOMINAL_LONGITUDE
        assert var.attrs["orbital_parameters"]["satellite_nominal_latitude"] == 0

    def test_times_are_in_dataset_attributes(self, nwcsaf_geo_ct_filehandler):
        """Check that start/end times are in the attributes of datasets."""
        dsid = {"name": "ct"}
        var = nwcsaf_geo_ct_filehandler.get_dataset(dsid, {})
        assert "start_time" in var.attrs
        assert "end_time" in var.attrs

    def test_start_time(self, nwcsaf_geo_ct_filehandler):
        """Test the start time property."""
        assert nwcsaf_geo_ct_filehandler.start_time == read_nwcsaf_time(NOMINAL_TIME)

    def test_end_time(self, nwcsaf_geo_ct_filehandler):
        """Test the end time property."""
        assert nwcsaf_geo_ct_filehandler.end_time == read_nwcsaf_time(END_TIME)

    def test_uint8_remains_uint8(self, nwcsaf_geo_ct_filehandler):
        """Test that loading uint8 remains uint8."""
        ct = nwcsaf_geo_ct_filehandler.get_dataset(
                {"name": "ct"},
                {"name": "ct", "file_type": "nc_nwcsaf_geo"})
        assert ct.dtype == np.dtype("uint8")

    def test_v2025_dimensions(self, nwcsaf_geo_v2025_ct_filehandler):
        """Test that the filehandler drops the time dimension for v2025 format."""
        dsid ={"name": "ct"}
        ds_info = {}
        data = nwcsaf_geo_v2025_ct_filehandler.get_dataset(dsid, ds_info)
        assert data.dims == ("y", "x")


class TestNcNWCSAFPPS:
    """Test the NcNWCSAF reader for PPS products."""

    def test_start_time(self, nwcsaf_pps_cmic_filehandler):
        """Test the start time property."""
        assert nwcsaf_pps_cmic_filehandler.start_time == read_nwcsaf_time(START_TIME_PPS)

    def test_end_time(self, nwcsaf_pps_cmic_filehandler):
        """Test the start time property."""
        assert nwcsaf_pps_cmic_filehandler.end_time == read_nwcsaf_time(END_TIME_PPS)

    def test_drop_xycoords(self, nwcsaf_pps_cmic_filehandler):
        """Test the drop of x and y coords."""
        y_line = xr.DataArray(list(range(5)), dims=("y"), attrs={"long_name": "scan line number"})
        x_pixel = xr.DataArray(list(range(10)), dims=("x"), attrs={"long_name": "pixel number"})
        lat = xr.DataArray(np.ones((5, 10)),
                           dims=("y", "x"),
                           coords={"y": y_line, "x": x_pixel},
                           attrs={"name": "lat",
                                  "standard_name": "latitude"})
        lon = xr.DataArray(np.ones((5, 10)),
                           dims=("y", "x"),
                           coords={"y": y_line, "x": x_pixel},
                           attrs={"name": "lon",
                                  "standard_name": "longitude"})
        data_array_in = xr.DataArray(np.ones((5, 10)),
                                     attrs={"scale_factor": np.array(0, dtype=float),
                                            "add_offset": np.array(1, dtype=float)},
                                     dims=("y", "x"),
                                     coords={"lon": lon, "lat": lat, "y": y_line, "x": x_pixel})
        data_array_out = nwcsaf_pps_cmic_filehandler.drop_xycoords(data_array_in)
        assert "y" not in data_array_out.coords

    def test_get_dataset_scales_and_offsets(self, nwcsaf_pps_cpp_filehandler):
        """Test that get_dataset() returns scaled and offseted data."""
        dsid = {"name": "cpp_cot"}

        info = dict(name="cpp_cot",
                    file_type="nc_nwcsaf_cpp")

        res = nwcsaf_pps_cpp_filehandler.get_dataset(dsid, info)
        np.testing.assert_allclose(res, COT_ARRAY * COT_SCALE + COT_OFFSET)

    def test_get_dataset_scales_and_offsets_palette_meanings_using_other_dataset(self, nwcsaf_pps_cpp_filehandler):
        """Test that get_dataset() returns scaled palette_meanings with another dataset as scaling source."""
        dsid = {"name": "cpp_cot_pal"}

        info = dict(name="cpp_cot_pal",
                    file_type="nc_nwcsaf_cpp",
                    scale_offset_dataset="cot")

        res = nwcsaf_pps_cpp_filehandler.get_dataset(dsid, info)
        palette_meanings = np.array(COT_PALETTE_MEANINGS.split()).astype(int)
        np.testing.assert_allclose(res.attrs["palette_meanings"], palette_meanings * COT_SCALE + COT_OFFSET)

    def test_get_palette_fill_value_color_added(self, nwcsaf_pps_ctth_filehandler):
        """Test that get_dataset() returns scaled palette_meanings with fill_value_color added."""
        dsid = {"name": "ctth_alti_pal"}

        info = dict(name="ctth_alti_pal",
                    file_type="nc_nwcsaf_ctth",
                    scale_offset_dataset="ctth_alti")

        res = nwcsaf_pps_ctth_filehandler.get_dataset(dsid, info)
        res.attrs["palette_meanings"]
        palette_meanings = np.array([0, 500, 1000, 1500, 65535])
        np.testing.assert_allclose(res.attrs["palette_meanings"], palette_meanings * COT_SCALE + COT_OFFSET)

    def test_get_dataset_raises_when_dataset_missing(self, nwcsaf_pps_cpp_filehandler):
        """Test that get_dataset() raises an error when the requested dataset is missing."""
        dsid = {"name": "cpp_phase"}
        info = dict(name="cpp_phase",
                    file_type="nc_nwcsaf_cpp")
        with pytest.raises(KeyError):
            nwcsaf_pps_cpp_filehandler.get_dataset(dsid, info)

    def test_get_dataset_uses_file_key_if_present(self, nwcsaf_pps_cmic_filehandler, nwcsaf_pps_cpp_filehandler):
        """Test that get_dataset() uses a file_key if present."""
        dsid_cpp = {"name": "cpp_cot"}
        dsid_cmic = {"name": "cmic_cot"}

        file_key = "cmic_cot"

        nwcsaf_pps_cmic_filehandler.file_key_prefix = ""

        info_cpp = dict(name="cpp_cot",
                        file_key=file_key,
                        file_type="nc_nwcsaf_cpp")

        res_cpp = nwcsaf_pps_cmic_filehandler.get_dataset(dsid_cpp, info_cpp)

        info_cmic = dict(name="cmic_cot",
                         file_type="nc_nwcsaf_cpp")

        res_cmic = nwcsaf_pps_cmic_filehandler.get_dataset(dsid_cmic, info_cmic)
        np.testing.assert_allclose(res_cpp, res_cmic)

    def test_get_dataset_can_handle_file_key_list(self, nwcsaf_pps_cmic_filehandler, nwcsaf_pps_cpp_filehandler):
        """Test that get_dataset() can handle a list of file_keys."""
        dsid_cpp = {"name": "cpp_reff"}
        dsid_cmic = {"name": "cmic_cre"}

        info_cpp = dict(name="cmic_reff",
                        file_key=["reff", "cre"],
                        file_type="nc_nwcsaf_cpp")

        res_cpp = nwcsaf_pps_cpp_filehandler.get_dataset(dsid_cpp, info_cpp)

        info_cmic = dict(name="cmic_reff",
                         file_key=["reff", "cre"],
                         file_type="nc_nwcsaf_cpp")

        res_cmic = nwcsaf_pps_cmic_filehandler.get_dataset(dsid_cmic, info_cmic)
        np.testing.assert_allclose(res_cpp, res_cmic)


class TestNcNWCSAFFileKeyPrefix:
    """Test the NcNWCSAF reader when using a file key prefix."""

    def test_get_dataset_uses_file_key_prefix(self, nwcsaf_pps_cmic_filehandler):
        """Test that get_dataset() uses a file_key_prefix."""
        dsid_cpp = {"name": "cpp_cot"}
        dsid_cmic = {"name": "cmic_cot"}

        file_key = "cot"

        info_cpp = dict(name="cpp_cot",
                        file_key=file_key,
                        file_type="nc_nwcsaf_cpp")

        res_cpp = nwcsaf_pps_cmic_filehandler.get_dataset(dsid_cpp, info_cpp)

        info_cmic = dict(name="cmic_cot",
                         file_type="nc_nwcsaf_cpp")

        res_cmic = nwcsaf_pps_cmic_filehandler.get_dataset(dsid_cmic, info_cmic)
        np.testing.assert_allclose(res_cpp, res_cmic)

    def test_get_dataset_scales_and_offsets_palette_meanings_using_other_dataset(self, nwcsaf_pps_cmic_filehandler):
        """Test that get_dataset() returns scaled palette_meanings using another dataset as scaling source."""
        dsid = {"name": "cpp_cot_pal"}

        info = dict(name="cpp_cot_pal",
                    file_key="cot_pal",
                    file_type="nc_nwcsaf_cpp",
                    scale_offset_dataset="cot")

        res = nwcsaf_pps_cmic_filehandler.get_dataset(dsid, info)
        palette_meanings = np.array(COT_PALETTE_MEANINGS.split()).astype(int)
        np.testing.assert_allclose(res.attrs["palette_meanings"], palette_meanings * COT_SCALE + COT_OFFSET)


def _check_filehandler_area_def(file_handler, dsid):
    from pyproj import CRS

    area_definition = file_handler.get_area_def(dsid)

    expected_crs = CRS(PROJ["gdal_projection"])
    assert area_definition.crs == expected_crs

    correct_extent = (PROJ["gdal_xgeo_up_left"],
                      PROJ["gdal_ygeo_low_right"],
                      PROJ["gdal_xgeo_low_right"],
                      PROJ["gdal_ygeo_up_left"])
    assert area_definition.area_extent == correct_extent



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the FY-3D MERSI-2 L1B file format.

The files for this reader are HDF5 and come in four varieties; band data
and geolocation data, both at 250m and 1000m resolution.

This reader was tested on FY-3A/B/C MERSI-1, FY-3D MERSI-2, FY-3E MERSI-LL and FY-3G MERSI-RM data,
but should work on future platforms as well assuming no file format changes.

"""

import datetime as dt

import dask.array as da
import numpy as np
from pyspectral.blackbody import blackbody_wn_rad2temp as rad2temp

from satpy.readers.core.hdf5 import HDF5FileHandler

N_TOT_IR_CHANS_LL = 6
PLATFORMS_INSTRUMENTS = {"FY-3A": "mersi-1",
                         "FY-3B": "mersi-1",
                         "FY-3C": "mersi-1",
                         "FY-3D": "mersi-2",
                         "FY-3E": "mersi-ll",
                         "FY-3F": "mersi-3",
                         "FY-3G": "mersi-rm"}


class MERSIL1B(HDF5FileHandler):
    """MERSI-1/MERSI-2/MERSI-LL/MERSI-RM L1B file reader."""

    def _strptime(self, date_attr, time_attr):
        """Parse date/time strings."""
        date = self[date_attr]
        time = self[time_attr]  # "18:27:39.720"
        # cuts off microseconds because of unknown meaning
        # is .720 == 720 microseconds or 720000 microseconds
        return dt.datetime.strptime(date + " " + time.split(".")[0], "%Y-%m-%d %H:%M:%S")

    @property
    def start_time(self):
        """Time for first observation."""
        return self._strptime("/attr/Observing Beginning Date", "/attr/Observing Beginning Time")

    @property
    def end_time(self):
        """Time for final observation."""
        return self._strptime("/attr/Observing Ending Date", "/attr/Observing Ending Time")

    @property
    def sensor_name(self):
        """Map sensor name to Satpy 'standard' sensor names."""
        return PLATFORMS_INSTRUMENTS.get(self.platform_name)

    @property
    def platform_name(self):
        """Platform name."""
        return self["/attr/Satellite Name"]

    def get_refl_mult(self):
        """Get reflectance multiplier."""
        if self.sensor_name == "mersi-rm":
            # MERSI-RM has reflectance in the range 0-1, so we need to convert
            return 100.
        else:
            return 1.

    def _get_single_slope_intercept(self, slope, intercept, cal_index):
        try:
            # convert scalar arrays to scalar
            return slope.item(), intercept.item()
        except ValueError:
            # numpy array but has more than one element
            return slope[cal_index], intercept[cal_index]

    def _get_coefficients(self, cal_key, cal_index):
        """Get VIS calibration coeffs from calibration datasets."""
        # Only one VIS band for MERSI-LL
        coeffs = self[cal_key][cal_index] if self.sensor_name != "mersi-ll" else self[cal_key]
        slope = coeffs.attrs.pop("Slope", None)
        intercept = coeffs.attrs.pop("Intercept", None)
        if slope is not None:
            slope, intercept = self._get_single_slope_intercept(
                slope, intercept, cal_index)
            coeffs = coeffs * slope + intercept
        return coeffs

    def _get_coefficients_mersi1(self, cal_index):
        """Get VIS calibration coeffs from attributes. Only for MERSI-1 on FY-3A/B."""
        try:
            # This is found in the actual file.
            coeffs = self["/attr/VIR_Cal_Coeff"]
        except KeyError:
            # This is in the official manual.
            coeffs = self["/attr/VIS_Cal_Coeff"]
        coeffs = coeffs.reshape(19, 3)
        coeffs = coeffs[cal_index].tolist()
        return coeffs

    def _get_dn_corrections(self, data, band_index, dataset_id, attrs):
        """Use slope and intercept to get DN corrections."""
        slope = attrs.pop("Slope", None)
        intercept = attrs.pop("Intercept", None)
        if slope is not None and dataset_id.get("calibration") != "counts":
            if band_index is not None and slope.size > 1:
                slope = slope[band_index]
                intercept = intercept[band_index]
            # There's a bug in slope for MERSI-1 IR band
            slope = 0.01 if self.sensor_name == "mersi-1" and dataset_id["name"] == "5" else slope
            data = data * slope + intercept
        return data

    def get_dataset(self, dataset_id, ds_info):
        """Load data variable and metadata and calibrate if needed."""
        file_key = ds_info.get("file_key", dataset_id["name"])
        band_index = ds_info.get("band_index")
        data = self[file_key]
        data = data[band_index] if band_index is not None else data
        data = data.rename({data.dims[-2]: "y", data.dims[-1]: "x"}) if data.ndim >= 2 else data

        attrs = data.attrs.copy()  # avoid contaminating other band loading
        attrs.update(ds_info)
        if "rows_per_scan" in self.filetype_info:
            attrs.setdefault("rows_per_scan", self.filetype_info["rows_per_scan"])

        data = self._mask_data(data, dataset_id, attrs)
        data = self._get_dn_corrections(data, band_index, dataset_id, attrs)

        if dataset_id.get("calibration") == "reflectance":
            data = self._get_ref_dataset(data, ds_info)

        elif dataset_id.get("calibration") == "radiance":
            data = self._get_rad_dataset(data, ds_info, dataset_id)

        elif dataset_id.get("calibration") == "brightness_temperature":
            # Converts um^-1 (wavenumbers) and (mW/m^2)/(str/cm^-1) (radiance data)
            # to SI units m^-1, mW*m^-3*str^-1.
            wave_number = 1. / (dataset_id["wavelength"][1] / 1e6)
            # MERSI-1 doesn't have additional corrections
            calibration_index = None if self.sensor_name == "mersi-1" else ds_info["calibration_index"]
            data = self._get_bt_dataset(data, calibration_index, wave_number)

        data.attrs = attrs
        # convert bytes to str
        for key, val in attrs.items():
            # python 3 only
            if bytes is not str and isinstance(val, bytes):
                data.attrs[key] = val.decode("utf8")

        data.attrs.update({
            "platform_name": self.platform_name,
            "sensor": self.sensor_name,
        })

        return data

    def _mask_data(self, data, dataset_id, attrs):
        """Mask the data using fill_value and valid_range attributes."""
        fill_value = attrs.pop("_FillValue", np.nan) if self.platform_name in ["FY-3A", "FY-3B"] else \
            attrs.pop("FillValue", np.nan) # covered by valid_range
        valid_range = attrs.pop("valid_range", None)
        if dataset_id.get("calibration") == "counts":
            # preserve integer type of counts if possible
            attrs["_FillValue"] = fill_value
            new_fill = data.dtype.type(fill_value)
        else:
            new_fill = np.float32(np.nan)
        try:
            # Due to a bug in the valid_range upper limit in the 10.8(24) and 12.0(25)
            # in the HDF data, this is hardcoded here.
            valid_range[1] = 25000 if self.sensor_name == "mersi-2" and dataset_id["name"] in ["24", "25"] and \
                                      valid_range[1] == 4095 else valid_range[1]
            # Similar bug also found in MERSI-1
            valid_range[1] = 25000 if self.sensor_name == "mersi-1" and dataset_id["name"] == "5" and \
                                      valid_range[1] == 4095 else valid_range[1]
            # typically bad_values == 65535, saturated == 65534
            # dead detector == 65533
            data = data.where((data >= valid_range[0]) &
                              (data <= valid_range[1]), new_fill)
            return data
        # valid_range could be None
        except TypeError:
            return data

    def _get_ref_dataset(self, data, ds_info):
        """Get the dataset as reflectance.

        For MERSI-1/2/RM, coefficients will be as::

            Reflectance = coeffs_1 + coeffs_2 * DN + coeffs_3 * DN ** 2

        For MERSI-LL, the DN value is in radiance and the reflectance could be calculated by::

            Reflectance = Rad * pi / E0 * 100

        Here E0 represents the solar irradiance of the specific band and is the coefficient.

        """
        # Only FY-3A/B stores VIS calibration coefficients in attributes
        coeffs = self._get_coefficients_mersi1(ds_info["calibration_index"]) if self.platform_name in ["FY-3A",
            "FY-3B"] else self._get_coefficients(ds_info["calibration_key"], ds_info.get("calibration_index", None))
        data = coeffs[0] + coeffs[1] * data + coeffs[2] * data ** 2 if self.sensor_name != "mersi-ll" else \
            data * np.pi / coeffs[0] * 100

        data = data * self.get_refl_mult()
        return data

    def _get_rad_dataset(self, data, ds_info, datset_id):
        """Get the dataset as radiance.

        For MERSI-2/RM VIS bands, this could be calculated by::

            Rad = Reflectance / 100 * E0 / pi

        For MERSI-2, E0 is in the attribute "Solar_Irradiance".
        For MERSI-RM, E0 is in the calibration dataset "Solar_Irradiance".
        However we can't find the way to retrieve this value from MERSI-1.

        For MERSI-LL VIS band, it has already been stored in DN values.
        After applying slope and intercept, we just get it. And Same way for IR bands, no matter which sensor it is.

        """
        mersi_2_vis = [str(i) for i in range(1, 20)]
        mersi_rm_vis = [str(i) for i in range(1, 6)]

        if self.sensor_name == "mersi-2" and datset_id["name"] in mersi_2_vis:
            E0 = self["/attr/Solar_Irradiance"]
            rad = self._get_ref_dataset(data, ds_info) / 100 * E0[mersi_2_vis.index(datset_id["name"])] / np.pi
        elif self.sensor_name == "mersi-rm" and datset_id["name"] in mersi_rm_vis:
            E0 = self._get_coefficients("Calibration/Solar_Irradiance", mersi_rm_vis.index(datset_id["name"]))
            rad = self._get_ref_dataset(data, ds_info) / 100 * E0 / np.pi
        else:
            rad = data
        return rad

    def _get_bt_dataset(self, data, calibration_index, wave_number):
        """Get the dataset as brightness temperature.

        Apparently we don't use these calibration factors for Rad -> BT::

            coeffs = self._get_coefficients(ds_info['calibration_key'], calibration_index)
            # coefficients are per-scan, we need to repeat the values for a
            # clean alignment
            coeffs = np.repeat(coeffs, data.shape[0] // coeffs.shape[1], axis=1)
            coeffs = coeffs.rename({
                coeffs.dims[0]: 'coefficients', coeffs.dims[1]: 'y'
            })  # match data dims
            data = coeffs[0] + coeffs[1] * data + coeffs[2] * data**2 + coeffs[3] * data**3

        """
        # pass the dask array
        bt_data = rad2temp(wave_number, data.data * 1e-5)  # brightness temperature

        # old versions of pyspectral produce numpy arrays
        # new versions of pyspectral can do dask arrays
        data.data = da.from_array(bt_data, chunks=data.data.chunks) if isinstance(bt_data, np.ndarray) else bt_data

        # Some BT bands seem to have 0 in the first 10 columns
        # and it is an invalid measurement, so let's mask
        data = data.where(data != 0)

        # additional corrections from the file
        if self.sensor_name == "mersi-1":
        # https://img.nsmc.org.cn/PORTAL/NSMC/DATASERVICE/SRF/FY3C/FY3C_MERSI_SRF.rar
            corr_coeff_a = 1.0047
            corr_coeff_b = -0.8549
        elif self.sensor_name == "mersi-2":
            corr_coeff_a = float(self["/attr/TBB_Trans_Coefficient_A"][calibration_index])
            corr_coeff_b = float(self["/attr/TBB_Trans_Coefficient_B"][calibration_index])
        elif self.sensor_name == "mersi-ll":
            # MERSI-LL stores these coefficients differently
            try:
                coeffs = self["/attr/TBB_Trans_Coefficient"]
                corr_coeff_a = coeffs[calibration_index]
                corr_coeff_b = coeffs[calibration_index + N_TOT_IR_CHANS_LL]
            except KeyError:
                return data
        else:
            # MERSI-RM has no correction coefficients
            corr_coeff_a = 0

        if corr_coeff_a != 0:
            data = (data - corr_coeff_b) / corr_coeff_a if self.sensor_name != "mersi-1" else \
                data * corr_coeff_a + corr_coeff_b
        # some bands have 0 counts for the first N columns and
        # seem to be invalid data points
        data = data.where(data != 0)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to MTG-FCI L1c NetCDF files.

This module defines the :class:`FCIL1cNCFileHandler` file handler, to
be used for reading Meteosat Third Generation (MTG) Flexible Combined
Imager (FCI) Level-1c data.  FCI flies
on the MTG Imager (MTG-I) series of satellites, with the first satellite (MTG-I1)
launched on the 13th of December 2022.
For more information about FCI, see `EUMETSAT`_.

To download data to be used with this reader, see `MTG data store guide`_.
For the Product User Guide (PUG) of the FCI L1c data, see `PUG`_.

Supported Data Variants
=======================

This reader currently supports Full Disk High Spectral Resolution Imagery
(FDHSI), High Spatial Resolution Fast Imagery (HRFI) data in full-disc ("FD") or in RSS ("Q4") scanning mode.
In addition, it also supports the L1C format for the African dissemination ("AF"), where each file
contains the masked full-disc of a single channel see `AF PUG`_.
Experimental support for special scans, e.g. with coverage "xx", is also given.

This reader supports data from both IDPF-I and IQT-I processing facilities.

.. note::

    If the user provides a list of both FDHSI and HRFI files from the same repeat cycle to the Satpy ``Scene``,
    Satpy will automatically read the channels from the source with the finest resolution,
    i.e. from the HRFI files for the vis_06, nir_22, ir_38, and ir_105 channels.
    If needed, the desired resolution can be explicitly requested using e.g.:
    ``scn.load(['vis_06'], resolution=1000)``.

Compressed Data
---------------

For reading compressed data, a decompression library is
needed. Either install the FCIDECOMP library (see `PUG`_), or the
``hdf5plugin`` package with::

    pip install hdf5plugin

or::

    conda install hdf5plugin -c conda-forge

.. note::

    If you use ``hdf5plugin``, make sure to add the line ``import hdf5plugin``
    at the top of your script.

Here's a common error when trying to read compressed data without one of
these plugins::

    RuntimeError: NetCDF: Filter error: undefined filter encountered


Geolocation
===========

Geolocation is based on information from the data files.  It uses:

    * From the shape of the data variable ``data/<channel>/measured/effective_radiance``,
      start and end line columns of current swath.
    * From the data variable ``data/<channel>/measured/x``, the x-coordinates
      for the grid, in radians (azimuth angle positive towards West).
    * From the data variable ``data/<channel>/measured/y``, the y-coordinates
      for the grid, in radians (elevation angle positive towards North).
    * From the attribute ``semi_major_axis`` on the data variable
      ``data/mtg_geos_projection``, the Earth equatorial radius
    * From the attribute ``inverse_flattening`` on the same data variable, the
      (inverse) flattening of the ellipsoid
    * From the attribute ``perspective_point_height`` on the same data
      variable, the geostationary altitude in the normalised geostationary
      projection
    * From the attribute ``longitude_of_projection_origin`` on the same
      data variable, the longitude of the projection origin
    * From the attribute ``sweep_angle_axis`` on the same, the sweep angle
      axis, see https://proj.org/operations/projections/geos.html

From the pixel centre angles in radians and the geostationary altitude, the
extremities of the lower left and upper right corners are calculated in units
of arc length in m.  This extent along with the number of columns and rows, the
sweep angle axis, and a dictionary with equatorial radius, polar radius,
geostationary altitude, and longitude of projection origin, are passed on to
``pyresample.geometry.AreaDefinition``, which then uses proj4 for the actual
geolocation calculations.

Image Orientation
-----------------

FCI scans south to north so that the native image orientation is "upside down"
(south up). By default Satpy preserves that orientation. You can flip the
image using the ``upper_right_corner`` keyword argument when loading datasets
(see :ref:`faq` for details).

.. code-block:: python

    scene.load(my_channels, upper_right_corner="NE")


Calibration
===========

The reading routine supports channel data in counts, radiances, and (depending
on channel) brightness temperatures or reflectances. The brightness temperature and reflectance calculation is based on
the formulas indicated in `PUG`_.
Radiance datasets are returned in units of radiance per unit wavenumber (mW m-2 sr-1 (cm-1)-1). Radiances can be
converted to units of radiance per unit wavelength (W m-2 um-1 sr-1) by multiplying with the
`radiance_unit_conversion_coefficient` dataset attribute.

Auxiliary Datasets
==================

For each channel, the reader also supports a number of auxiliary datasets, such as the pixel quality,
the index map and the related geometric and acquisition parameters: time,
subsatellite latitude, subsatellite longitude, platform altitude, subsolar latitude, subsolar longitude,
earth-sun distance, sun-satellite distance,  swath number, and swath direction.

All auxiliary data can be obtained by prepending the channel name such as
``"vis_04_pixel_quality"``.

.. warning::
    The API for the direct reading of pixel quality is temporary and likely to
    change.  Currently, for each channel, the pixel quality is available by
    ``<chan>_pixel_quality``.  In the future, they will likely all be called
    ``pixel_quality`` and disambiguated by a to-be-decided property in the
    `DataID`.

Memory Footprint
================

The FCI L1C reader has been optimized to load multiple channels as fast as
possible. This involves a lot of caching, which takes some extra time for Scene
creation and also increases the amount of memory allocated by the Scene itself,
even without loading any datasets.

The penalty is largest for single channel workflows. In this case it can help
to compute the data as soon as possible and delete the corresponding scene
object.

.. code-block:: python

    import gc

    channel_data = scene["ir_105"].compute()

    # Delete original scene to release memory. Run garbage collector to prevent
    # occasional segmentation faults, see
    # https://github.com/pytroll/satpy/issues/3144
    del scene
    gc.collect()

The number of dask workers also has an impact on the memory footprint, see :ref:`faq`.


.. _AF PUG: https://user.eumetsat.int/resources/user-guides/mtg-africa-data-service-guide  # noqa: E501
.. _PUG: https://user.eumetsat.int/resources/user-guides/mtg-fci-level-1c-data-guide  # noqa: E501
.. _EUMETSAT: https://user.eumetsat.int/resources/user-guides/mtg-fci-level-1c-data-guide  # noqa: E501
.. _MTG data store guide: https://user.eumetsat.int/resources/user-guides/data-store-mtg-data-access-guide  # noqa: E501
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import datetime as dt
import logging
from functools import cached_property

import dask.array as da
import numpy as np
import xarray as xr
from netCDF4 import default_fillvals
from pyorbital.astronomy import sun_earth_distance_correction
from pyresample import geometry

import satpy
from satpy.readers.core._geos_area import get_geos_area_naming
from satpy.readers.core.eum import get_service_mode
from satpy.readers.core.fci import platform_name_translate
from satpy.readers.core.netcdf import NetCDF4FsspecFileHandler

logger = logging.getLogger(__name__)

# dict containing all available auxiliary data parameters to be read using the index map. Keys are the
# parameter name and values are the paths to the variable inside the netcdf
AUX_DATA = {
    "subsatellite_latitude": "state/platform/subsatellite_latitude",
    "subsatellite_longitude": "state/platform/subsatellite_longitude",
    "platform_altitude": "state/platform/platform_altitude",
    "subsolar_latitude": "state/celestial/subsolar_latitude",
    "subsolar_longitude": "state/celestial/subsolar_longitude",
    "earth_sun_distance": "state/celestial/earth_sun_distance",
    "sun_satellite_distance": "state/celestial/sun_satellite_distance",
    "time": "time",
    "swath_number": "data/swath_number",
    "swath_direction": "data/swath_direction",
}

HIGH_RES_GRID_INFO = {"fci_l1c_hrfi": {"grid_type": "500m",
                                       "grid_width": 22272},
                      "fci_l1c_fdhsi": {"grid_type": "1km",
                                        "grid_width": 11136},
                      }
LOW_RES_GRID_INFO = {"fci_l1c_hrfi": {"grid_type": "1km",
                                      "grid_width": 11136},
                     "fci_l1c_fdhsi": {"grid_type": "2km",
                                       "grid_width": 5568},
                     }


def _get_aux_data_name_from_dsname(dsname):
    aux_data_name = [key for key in AUX_DATA.keys() if key in dsname]
    if len(aux_data_name) > 0:
        return aux_data_name[0]

    return None


def _get_channel_name_from_dsname(dsname):
    # FIXME: replace by .removesuffix after we drop support for Python < 3.9
    if dsname.endswith("_pixel_quality"):
        channel_name = dsname[:-len("_pixel_quality")]
    elif dsname.endswith("_index_map"):
        channel_name = dsname[:-len("_index_map")]
    elif _get_aux_data_name_from_dsname(dsname) is not None:
        channel_name = dsname[:-len(_get_aux_data_name_from_dsname(dsname)) - 1]
    else:
        channel_name = dsname

    return channel_name


class FCIL1cNCFileHandler(NetCDF4FsspecFileHandler):
    """Class implementing the MTG FCI L1c Filehandler.

    This class implements the Meteosat Third Generation (MTG) Flexible
    Combined Imager (FCI) Level-1c NetCDF reader.
    It is designed to be used through the :class:`satpy.Scene <satpy.scene.Scene>`
    class using the :mod:`Scene.load <satpy.scene.Scene.load>` method with the reader
    ``"fci_l1c_nc"``.

    """
    def __init__(self, filename, filename_info, filetype_info,
                 clip_negative_radiances=None, **kwargs):
        """Initialize file handler."""
        super().__init__(filename, filename_info,
                         filetype_info,
                         cache_var_size=0,
                         cache_handle=True)
        logger.debug("Reading: {}".format(self.filename))
        logger.debug("Start: {}".format(self.start_time))
        logger.debug("End: {}".format(self.end_time))

        if self.filename_info["coverage"] == "Q4":
            # change the chunk number so that padding gets activated correctly for Q4, which corresponds to the upper
            # quarter of the disc
            self.filename_info["count_in_repeat_cycle"] += 28

        if self.filename_info["coverage"] == "AF":
            # change number of chunk from 0 to 1 so that the padding is not activated (chunk 1 is present and only 1
            # chunk is expected), as the African dissemination products come in one file per full disk.
            self.filename_info["count_in_repeat_cycle"] = 1

        if self.filename_info["coverage"] == "xx":
            # add one to the chunk numbering to activate the padding mechanism for special scans (expected to have less
            # than 40 chunks, but still starting with 1)
            self.filename_info["count_in_repeat_cycle"] += 1

        if self.filename_info["facility_or_tool"] == "IQTI":
            self.is_iqt = True
        else:
            self.is_iqt = False

        if clip_negative_radiances is None:
            clip_negative_radiances = satpy.config.get("readers.clip_negative_radiances")
        self.clip_negative_radiances = clip_negative_radiances
        self._cache = {}

    @property
    def rc_period_min(self):
        """Get nominal repeat cycle duration."""
        if "Q4" in self.filename_info["coverage"]:
            return 2.5
        elif self.filename_info["coverage"] in ["FD", "AF"]:
            return 10
        else:
            logger.debug(f"Coverage \"{self.filename_info['coverage']}\" not recognised. "
                         f"Using observation times for nominal times.")
            return None


    @property
    def nominal_start_time(self):
        """Get nominal start time."""
        if self.rc_period_min is None:
            return self.filename_info["start_time"]
        else:
            rc_date = self.observation_start_time.replace(hour=0, minute=0, second=0, microsecond=0)
            return rc_date + dt.timedelta(
                minutes=(self.filename_info["repeat_cycle_in_day"] - 1) * self.rc_period_min)

    @property
    def nominal_end_time(self):
        """Get nominal end time."""
        if self.rc_period_min is None:
            return self.filename_info["end_time"]
        else:
            return self.nominal_start_time + dt.timedelta(minutes=self.rc_period_min)

    @property
    def observation_start_time(self):
        """Get observation start time."""
        return self.filename_info["start_time"]

    @property
    def observation_end_time(self):
        """Get observation end time."""
        return self.filename_info["end_time"]

    @property
    def start_time(self):
        """Get start time."""
        return self.nominal_start_time

    @property
    def end_time(self):
        """Get end time."""
        return self.nominal_end_time

    def get_channel_measured_group_path(self, channel):
        """Get the channel's measured group path."""
        if self.filetype_info["file_type"] == "fci_l1c_hrfi":
            channel += "_hr"
        measured_group_path = "data/{}/measured".format(channel)

        return measured_group_path

    def get_segment_position_info(self):
        """Get information about the size and the position of the segment inside the final image array.

        As the final array is composed by stacking segments vertically, the position of a segment
        inside the array is defined by the numbers of the start (lowest) and end (highest) row of the segment.
        The row numbering is assumed to start with 1.
        This info is used in the GEOVariableSegmentYAMLReader to compute optimal segment sizes for missing segments.

        Note: in the FCI terminology, a segment is actually called "chunk". To avoid confusion with the dask concept
        of chunk, and to be consistent with SEVIRI, we opt to use the word segment.

        Note: This function is not used for the African data as it contains only one segment.
        """
        file_type = self.filetype_info["file_type"]
        vis_06_measured_path = self.get_channel_measured_group_path("vis_06")
        ir_105_measured_path = self.get_channel_measured_group_path("ir_105")
        segment_position_info = {
            HIGH_RES_GRID_INFO[file_type]["grid_type"]: {
                "start_position_row": self.get_and_cache_npxr(vis_06_measured_path + "/start_position_row").item(),
                "end_position_row": self.get_and_cache_npxr(vis_06_measured_path + "/end_position_row").item(),
                "segment_height": self.get_and_cache_npxr(vis_06_measured_path + "/end_position_row").item() -
                                  self.get_and_cache_npxr(vis_06_measured_path + "/start_position_row").item() + 1,
                "grid_width": HIGH_RES_GRID_INFO[file_type]["grid_width"]
            },
            LOW_RES_GRID_INFO[file_type]["grid_type"]: {
                "start_position_row": self.get_and_cache_npxr(ir_105_measured_path + "/start_position_row").item(),
                "end_position_row": self.get_and_cache_npxr(ir_105_measured_path + "/end_position_row").item(),
                "segment_height": self.get_and_cache_npxr(ir_105_measured_path + "/end_position_row").item() -
                                  self.get_and_cache_npxr(ir_105_measured_path + "/start_position_row").item() + 1,
                "grid_width": LOW_RES_GRID_INFO[file_type]["grid_width"]
            }
        }
        return segment_position_info

    def get_dataset(self, key, info=None):
        """Load a dataset."""
        logger.debug("Reading {} from {}".format(key["name"], self.filename))
        if "pixel_quality" in key["name"]:
            return self._get_dataset_quality(key, info=info)
        elif "index_map" in key["name"]:
            return self._get_dataset_index_map(key, info=info)
        elif _get_aux_data_name_from_dsname(key["name"]) is not None:
            return self._get_dataset_aux_data(key, info=info)
        elif any(lb in key["name"] for lb in {"vis_", "ir_", "nir_", "wv_"}):
            return self._get_dataset_measurand(key, info=info)
        else:
            raise ValueError("Unknown dataset key, not a channel, quality or auxiliary data: "
                             f"{key['name']:s}")

    def _get_dataset_measurand(self, key, info=None):
        """Load dataset corresponding to channel measurement.

        Load a dataset when the key refers to a measurand, whether uncalibrated
        (counts) or calibrated in terms of brightness temperature, radiance, or
        reflectance.
        """
        # Get the dataset
        # Get metadata for given dataset
        measured = self.get_channel_measured_group_path(key["name"])
        data = self[measured + "/effective_radiance"]

        attrs = dict(data.attrs).copy()
        info = info.copy()
        data = _ensure_dataarray(data)

        fv = attrs.pop(
            "_FillValue",
            default_fillvals.get(data.dtype.str[1:], np.float32(np.nan)))
        vr = attrs.get("valid_range", [np.float32(-np.inf), np.float32(np.inf)])
        if key["calibration"] == "counts":
            attrs["_FillValue"] = fv
            nfv = data.dtype.type(fv)
        else:
            nfv = np.float32(np.nan)
        data = data.where((data >= vr[0]) & (data <= vr[1]), nfv)

        res = self.calibrate(data, key)
        cleaned_attrs = self._set_and_cleanup_attributes(res.attrs, attrs, key, info)
        res.attrs = cleaned_attrs
        return res

    def _set_calibrated_data_attributes(self, resattrs, attrs, key):
        # For each channel, the effective_radiance contains in the
        # "ancillary_variables" attribute the value "pixel_quality".  In
        # FileYAMLReader._load_ancillary_variables, satpy will try to load
        # "pixel_quality" but is lacking the context from what group to load
        # it: in the FCI format, each channel group (data/<channel>/measured) has
        # its own data variable 'pixel_quality'.
        # Until we can have multiple pixel_quality variables defined (for
        # example, with https://github.com/pytroll/satpy/pull/1088), rewrite
        # the ancillary variable to include the channel. See also
        # https://github.com/pytroll/satpy/issues/1171.
        if "pixel_quality" in attrs["ancillary_variables"]:
            attrs["ancillary_variables"] = attrs["ancillary_variables"].replace(
                "pixel_quality", key["name"] + "_pixel_quality")
        else:
            raise ValueError(
                "Unexpected value for attribute ancillary_variables, "
                "which the FCI file handler intends to rewrite (see "
                "https://github.com/pytroll/satpy/issues/1171 for why). "
                f"Expected 'pixel_quality', got {attrs['ancillary_variables']:s}")

        resattrs.update(key.to_dict())

        # pre-calibration units no longer apply
        attrs.pop("units")
        resattrs.update(attrs)

        # remove unpacking parameters for calibrated data
        if key["calibration"] in ["brightness_temperature", "reflectance", "radiance"]:
            resattrs.pop("add_offset")
            resattrs.pop("warm_add_offset")
            resattrs.pop("scale_factor")
            resattrs.pop("warm_scale_factor")
            resattrs.pop("valid_range")
            # Some xarray versions can propagate _FillValue too:
            resattrs.pop("_FillValue", None)

        # remove attributes from original file which don't apply anymore
        resattrs.pop("long_name")
        return resattrs

    def _set_and_cleanup_attributes(self, resattrs, attrs=None, key=None, info=None):
        """Gather attributes for the result dataset and remove unwanted keys.

        Adds global attributes to all datasets.
        For measurand data, adds dataset-specific attributes.
        Ensures attributes from uncalibrated data (e.g. `_FillValue` from counts)
        are not propagated to the calibrated data.
        """
        if info is not None:
            resattrs.update(info)
        if None not in (attrs, key):
            resattrs = self._set_calibrated_data_attributes(resattrs, attrs, key)

        resattrs["platform_name"] = platform_name_translate.get(
            self["attr/platform"], self["attr/platform"])

        # Add time_parameter attributes
        resattrs["time_parameters"] = {
            "nominal_start_time": self.nominal_start_time,
            "nominal_end_time": self.nominal_end_time,
            "observation_start_time": self.observation_start_time,
            "observation_end_time": self.observation_end_time,
        }

        resattrs.update(self.orbital_param)
        return resattrs

    def get_iqt_parameters_lon_lat_alt(self):
        """Compute the orbital parameters for IQT data.

        Compute satellite_actual_longitude, satellite_actual_latitude, satellite_actual_altitude.
        """
        actual_subsat_lon = float(self.get_and_cache_npxr("data/mtg_geos_projection/attr/"
                                                          "longitude_of_projection_origin"))
        actual_subsat_lat = 0.0
        actual_sat_alt = float(self.get_and_cache_npxr("data/mtg_geos_projection/attr/perspective_point_height"))
        logger.info("For IQT data, the following parameter is hardcoded:"
                    f" satellite_actual_latitude = {actual_subsat_lat}. "
                    "The following parameters are taken from the projection dictionary: "
                    f"satellite_actual_longitude = {actual_subsat_lon}, "
                    f"satellite_actual_altitude = {actual_sat_alt}")
        return actual_subsat_lon, actual_subsat_lat, actual_sat_alt

    def get_parameters_lon_lat_alt(self):
        """Compute the orbital parameters.

        Compute satellite_actual_longitude, satellite_actual_latitude, satellite_actual_altitude.
        """
        actual_subsat_lon = float(np.nanmean(self._get_aux_data_lut_vector("subsatellite_longitude")))
        actual_subsat_lat = float(np.nanmean(self._get_aux_data_lut_vector("subsatellite_latitude")))
        actual_sat_alt = float(np.nanmean(self._get_aux_data_lut_vector("platform_altitude")))
        return actual_subsat_lon, actual_subsat_lat, actual_sat_alt

    @cached_property
    def orbital_param(self):
        """Compute the orbital parameters for the current segment."""
        if self.is_iqt:
            actual_subsat_lon, actual_subsat_lat, actual_sat_alt = self.get_iqt_parameters_lon_lat_alt()
        else:
            actual_subsat_lon, actual_subsat_lat, actual_sat_alt = self.get_parameters_lon_lat_alt()
        # The "try" is a temporary part of the code as long as the AF data are not fixed
        try:
            nominal_and_proj_subsat_lon = float(
                self.get_and_cache_npxr("data/mtg_geos_projection/attr/longitude_of_projection_origin"))
        except ValueError:
            nominal_and_proj_subsat_lon = 0.0
        nominal_and_proj_subsat_lat = 0.0
        nominal_and_proj_sat_alt = float(
            self.get_and_cache_npxr("data/mtg_geos_projection/attr/perspective_point_height"))

        orb_param_dict = {
            "orbital_parameters": {
                "satellite_actual_longitude": actual_subsat_lon,
                "satellite_actual_latitude": actual_subsat_lat,
                "satellite_actual_altitude": actual_sat_alt,
                "satellite_nominal_longitude": nominal_and_proj_subsat_lon,
                "satellite_nominal_latitude": nominal_and_proj_subsat_lat,
                "satellite_nominal_altitude": nominal_and_proj_sat_alt,
                "projection_longitude": nominal_and_proj_subsat_lon,
                "projection_latitude": nominal_and_proj_subsat_lat,
                "projection_altitude": nominal_and_proj_sat_alt,
            }}

        return orb_param_dict

    def _get_dataset_quality(self, key, info=None):
        """Load a quality field for an FCI channel."""
        grp_path = self.get_channel_measured_group_path(_get_channel_name_from_dsname(key["name"]))
        dv_path = grp_path + "/pixel_quality"
        data = self[dv_path]
        cleaned_attrs = self._set_and_cleanup_attributes(data.attrs, key=key, info=info)
        data.attrs = cleaned_attrs
        return data

    def _get_dataset_index_map(self, key, info=None):
        """Load the index map for an FCI channel."""
        dsname = key["name"]
        grp_path = self.get_channel_measured_group_path(_get_channel_name_from_dsname(dsname))
        dv_path = grp_path + "/index_map"
        data = self[dv_path]

        data = data.where(data != data.attrs.get("_FillValue", 65535))
        cleaned_attrs = self._set_and_cleanup_attributes(data.attrs, info=info)
        data.attrs = cleaned_attrs
        return data

    def _get_aux_data_lut_vector(self, aux_data_name):
        """Load the lut vector of an auxiliary variable."""
        lut = self.get_and_cache_npxr(AUX_DATA[aux_data_name])
        lut = _ensure_dataarray(lut)
        fv = default_fillvals.get(lut.dtype.str[1:], np.nan)
        lut = lut.where(lut != fv)

        return lut

    @staticmethod
    def _getitem(block, lut):
        return lut[block.astype("uint16")]

    def _get_dataset_aux_data(self, key, info=None):
        """Get the auxiliary data arrays using the index map."""
        dsname = key["name"]
        # get index map
        index_map = self._get_dataset_index_map(key)
        # subtract minimum of index variable (index_offset)
        index_map -= np.min(self.get_and_cache_npxr("index"))

        # get lut values from 1-d vector variable
        lut = self._get_aux_data_lut_vector(_get_aux_data_name_from_dsname(dsname))
        # assign lut values based on index map indices
        aux = index_map.data.map_blocks(self._getitem, lut.data, dtype=lut.data.dtype)
        aux = xr.DataArray(aux, dims=index_map.dims, attrs=index_map.attrs, coords=index_map.coords)

        # filter out out-of-disk values
        aux = aux.where(index_map >= 0)

        cleaned_attrs = self._set_and_cleanup_attributes(aux.attrs, key=key, info=info)
        aux.attrs = cleaned_attrs
        return aux

    def calc_area_extent(self, key):
        """Calculate area extent for a dataset."""
        # if a user requests a pixel quality or index map before the channel data, the
        # yaml-reader will ask the area extent of the pixel quality/index map field,
        # which will ultimately end up here
        channel_name = _get_channel_name_from_dsname(key["name"])
        # Get metadata for given dataset
        measured = self.get_channel_measured_group_path(channel_name)
        # Get start/end line and column of loaded swath.
        nlines, ncols = self[measured + "/effective_radiance/shape"]

        logger.debug("Channel {} resolution: {}".format(channel_name, ncols))
        logger.debug("Row/Cols: {} / {}".format(nlines, ncols))

        # Calculate full globe line extent
        h = float(self.get_and_cache_npxr("data/mtg_geos_projection/attr/perspective_point_height"))

        extents = {}
        for coord in "xy":
            coord_radian = self.get_and_cache_npxr(measured + "/{:s}".format(coord))

            coord_radian_num = coord_radian[:] * coord_radian.attrs["scale_factor"] + coord_radian.attrs["add_offset"]

            # FCI defines pixels by centroids (see PUG), while pyresample
            # defines corners as lower left corner of lower left pixel, upper right corner of upper right pixel
            # (see https://pyresample.readthedocs.io/en/latest/geo_def.html).
            # Therefore, half a pixel (i.e. half scale factor) needs to be added in each direction.

            # The grid origin is in the South-West corner.
            # Note that the azimuth angle (x) is defined as positive towards West (see PUG - Level 1c Reference Grid)
            # The elevation angle (y) is defined as positive towards North as per usual convention. Therefore:
            # The values of x go from positive (West) to negative (East) and the scale factor of x is negative.
            # The values of y go from negative (South) to positive (North) and the scale factor of y is positive.

            # South-West corner (x positive, y negative)
            first_coord_radian = coord_radian_num[0] - coord_radian.attrs["scale_factor"] / 2
            # North-East corner (x negative, y positive)
            last_coord_radian = coord_radian_num[-1] + coord_radian.attrs["scale_factor"] / 2

            # convert to arc length in m
            first_coord = first_coord_radian * h  # arc length in m
            last_coord = last_coord_radian * h

            # the .item() call is needed with the h5netcdf backend, see
            # https://github.com/pytroll/satpy/issues/972#issuecomment-558191583
            # but we need to compute it first if this is dask
            try:
                first_coord = first_coord.compute()
                last_coord = last_coord.compute()
            except AttributeError:  # not a dask.array
                pass

            extents[coord] = (first_coord.item(), last_coord.item())

        # For the final extents, take into account that the image is upside down (lower line is North), and that
        # East is defined as positive azimuth in Proj, so we need to multiply by -1 the azimuth extents.

        # lower left x: west-ward extent: first coord of x, multiplied by -1 to account for azimuth orientation
        # lower left y: north-ward extent: last coord of y
        # upper right x: east-ward extent: last coord of x, multiplied by -1 to account for azimuth orientation
        # upper right y: south-ward extent: first coord of y
        area_extent = (-extents["x"][0], extents["y"][1], -extents["x"][1], extents["y"][0])

        return area_extent, nlines, ncols

    def get_area_def(self, key):
        """Calculate on-fly area definition for a dataset in geos-projection."""
        # assumption: channels with same resolution should have same area
        # cache results to improve performance
        if key["resolution"] in self._cache:
            return self._cache[key["resolution"]]

        a = float(self.get_and_cache_npxr("data/mtg_geos_projection/attr/semi_major_axis"))
        h = float(self.get_and_cache_npxr("data/mtg_geos_projection/attr/perspective_point_height"))
        rf = float(self.get_and_cache_npxr("data/mtg_geos_projection/attr/inverse_flattening"))
        # The "try" is a temporary part of the code as long as the AF data are not modified
        try:
            lon_0 = float(self.get_and_cache_npxr("data/mtg_geos_projection/attr/longitude_of_projection_origin"))
        except ValueError:
            lon_0 = 0.0
        sweep = str(self.get_and_cache_npxr("data/mtg_geos_projection/attr/sweep_angle_axis"))

        area_extent, nlines, ncols = self.calc_area_extent(key)
        logger.debug("Calculated area extent: {}"
                     .format("".join(str(area_extent))))

        # use a (semi-major axis) and rf (reverse flattening) to define ellipsoid as recommended by EUM (see PUG)
        proj_dict = {"a": a,
                     "lon_0": lon_0,
                     "h": h,
                     "rf": rf,
                     "proj": "geos",
                     "units": "m",
                     "sweep": sweep}

        area_naming_input_dict = {"platform_name": "mtg",
                                  "instrument_name": "fci",
                                  "resolution": int(key["resolution"])
                                  }
        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode("fci", lon_0)})

        area = geometry.AreaDefinition(
            area_naming["area_id"],
            area_naming["description"],
            "",
            proj_dict,
            ncols,
            nlines,
            area_extent)

        self._cache[key["resolution"]] = area
        return area

    def calibrate(self, data, key):
        """Calibrate data."""
        if key["calibration"] in ["brightness_temperature", "reflectance", "radiance"]:
            data = self.calibrate_counts_to_physical_quantity(data, key)
        elif key["calibration"] != "counts":
            logger.error(
                "Received unknown calibration key.  Expected "
                "'brightness_temperature', 'reflectance', 'radiance' or 'counts', got "
                + key["calibration"] + ".")

        return data

    def calibrate_counts_to_physical_quantity(self, data, key):
        """Calibrate counts to radiances, brightness temperatures, or reflectances."""
        # counts to radiance scaling

        data = self.calibrate_counts_to_rad(data, key)

        if key["calibration"] == "brightness_temperature":
            data = self.calibrate_rad_to_bt(data, key)
        elif key["calibration"] == "reflectance":
            data = self.calibrate_rad_to_refl(data, key)

        return data

    def calibrate_counts_to_rad(self, data, key):
        """Calibrate counts to radiances."""
        if self.clip_negative_radiances:
            data = self._clipneg(data)
        if key["name"] == "ir_38":
            data = xr.where(((2 ** 12 - 1 < data) & (data <= 2 ** 13 - 1)),
                            (data * data.attrs.get("warm_scale_factor", 1) +
                             data.attrs.get("warm_add_offset", 0)),
                            (data * data.attrs.get("scale_factor", 1) +
                             data.attrs.get("add_offset", 0))
                            )
        else:
            data = (data * data.attrs.get("scale_factor", 1) +
                    data.attrs.get("add_offset", 0))

        measured = self.get_channel_measured_group_path(key["name"])
        data.attrs.update({"radiance_unit_conversion_coefficient":
                               self.get_and_cache_npxr(measured + "/radiance_unit_conversion_coefficient")})
        return data

    @staticmethod
    def _clipneg(data):
        """Clip counts to avoid negative radiances."""
        lo = -data.attrs.get("add_offset", 0) // data.attrs.get("scale_factor", 1) + 1
        return data.where((~data.notnull())|(data>=lo), lo)

    def calibrate_rad_to_bt(self, radiance, key):
        """IR channel calibration."""
        # using the method from PUG section Converting from Effective Radiance to Brightness Temperature for IR Channels
        measured = self.get_channel_measured_group_path(key["name"])

        vc = self.get_and_cache_npxr(measured + "/radiance_to_bt_conversion_coefficient_wavenumber").astype(np.float32)

        a = self.get_and_cache_npxr(measured + "/radiance_to_bt_conversion_coefficient_a").astype(np.float32)
        b = self.get_and_cache_npxr(measured + "/radiance_to_bt_conversion_coefficient_b").astype(np.float32)

        c1 = self.get_and_cache_npxr(measured + "/radiance_to_bt_conversion_constant_c1").astype(np.float32)
        c2 = self.get_and_cache_npxr(measured + "/radiance_to_bt_conversion_constant_c2").astype(np.float32)

        for v in (vc, a, b, c1, c2):
            if v == v.attrs.get("_FillValue",
                                default_fillvals.get(v.dtype.str[1:])):
                logger.error(
                    "{:s} set to fill value, cannot produce "
                    "brightness temperatures for {:s}.".format(
                        v.attrs.get("long_name",
                                    "at least one necessary coefficient"),
                        measured))
                return radiance * np.float32(np.nan)

        nom = c2 * vc
        denom = a * np.log(1 + (c1 * vc ** np.float32(3.)) / radiance)

        res = nom / denom - b / a
        return res

    def calibrate_rad_to_refl(self, radiance, key):
        """VIS channel calibration."""
        measured = self.get_channel_measured_group_path(key["name"])

        cesi = self.get_and_cache_npxr(measured + "/channel_effective_solar_irradiance").astype(np.float32)

        if cesi == cesi.attrs.get(
                "_FillValue", default_fillvals.get(cesi.dtype.str[1:])):
            logger.error(
                "channel effective solar irradiance set to fill value, "
                "cannot produce reflectance for {:s}.".format(measured))
            return radiance * np.float32(np.nan)
        sun_earth_distance = self._compute_sun_earth_distance
        res = 100 * radiance * np.float32(np.pi) * np.float32(sun_earth_distance) ** np.float32(2) / cesi
        return res

    @cached_property
    def _compute_sun_earth_distance(self) -> float:
        """Compute the sun_earth_distance."""
        if self.is_iqt:
            middle_time_diff = (self.observation_end_time - self.observation_start_time) / 2
            utc_date = self.observation_start_time + middle_time_diff
            sun_earth_distance = sun_earth_distance_correction(utc_date)
            logger.info(f"The value sun_earth_distance is set to {sun_earth_distance} AU.")
        else:
            sun_earth_distance = np.nanmean(
                self._get_aux_data_lut_vector("earth_sun_distance")) / 149597870.7  # [AU]
        return sun_earth_distance


def _ensure_dataarray(arr):
    if not isinstance(arr, xr.DataArray):
        attrs = dict(arr.attrs).copy()
        arr = xr.DataArray(da.from_array(arr), dims=arr.dimensions, attrs=attrs, name=arr.name)
    return arr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""EUMETSAT EPS-SG Visible/Infrared Imager (VII) readers base class."""


import datetime as dt
import logging

import xarray as xr
from geotiepoints.viiinterpolator import tie_points_geo_interpolation, tie_points_interpolation

from satpy.readers.core.netcdf import NetCDF4FileHandler
from satpy.readers.core.vii import SCAN_ALT_TIE_POINTS, TIE_POINTS_FACTOR

logger = logging.getLogger(__name__)


class ViiNCBaseFileHandler(NetCDF4FileHandler):
    """Base reader class for VII products in netCDF format.

    Args:
        filename (str): File to read
        filename_info (dict): Dictionary with filename information
        filetype_info (dict): Dictionary with filetype information
        orthorect (bool): activates the orthorectification correction where available

    """

    # In the future, only final format will be defined once operational data is available
    DATETIME_FORMATS = [
        "%Y%m%d%H%M%S.%f",        # e.g. 20250924121530.123456
        "%Y-%m-%dT%H:%M:%S.%f",   # e.g. 2025-09-24T12:15:30.123456
        "%Y-%m-%d %H:%M:%S",      # e.g. 2025-09-24 12:15:30
        "%Y-%m-%d %H:%M:%S.%f",   # e.g. 2025-09-24 12:15:30.123456
    ]

    def __init__(self, filename, filename_info, filetype_info, orthorect=False):
        """Prepare the class for dataset reading."""
        super().__init__(filename, filename_info, filetype_info, auto_maskandscale=True)

        # Saves the orthorectification flag
        self.orthorect = orthorect and filetype_info.get("orthorect", True)

        # Saves the interpolation flag
        self.interpolate = filetype_info.get("interpolate", True)

        try:
            longitude = self[filetype_info["cached_longitude"]]
            latitude = self[filetype_info["cached_latitude"]]

            if self.interpolate:
                self.longitude, self.latitude = self._perform_geo_interpolation(longitude, latitude)
            else:
                self.longitude, self.latitude = longitude, latitude

        except KeyError:
            logger.warning("Cached longitude and/or latitude datasets are not correctly defined in YAML file")
            self.longitude, self.latitude = None, None

    def _standardize_dims(self, variable):
        """Standardize dims to y, x."""
        if "num_pixels" in variable.dims:
            variable = variable.rename({"num_pixels": "x", "num_lines": "y"})
        if "num_points_act" in variable.dims:
            variable = variable.rename({"num_points_act": "x", "num_points_alt": "y"})
        if variable.dims[0] == "x":
            variable = variable.transpose("y", "x")
        return variable

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using file_key in dataset_info."""
        var_key = dataset_info["file_key"]
        logger.debug("Reading in file to get dataset with key %s.", var_key)

        if var_key == "cached_longitude" and self.longitude is not None:
            variable = self.longitude.copy()
        elif var_key == "cached_latitude" and self.latitude is not None:
            variable = self.latitude.copy()
        else:
            try:
                variable = self[var_key]
            except KeyError:
                logger.warning("Could not find key %s in NetCDF file, no valid Dataset created", var_key)
                return None

            # If the dataset is marked for interpolation, perform the interpolation from tie points to pixels
            if dataset_info.get("interpolate", False) and self.interpolate:
                variable = self._perform_interpolation(variable)

            # Perform the calibration if required
            if dataset_info.get("calibration") is not None:
                variable = self._perform_calibration(variable, dataset_info)

        # Perform the orthorectification if required
        if self.orthorect:
            orthorect_data_name = dataset_info.get("orthorect_data", None)
            if orthorect_data_name is not None:
                variable = self._perform_orthorectification(variable, orthorect_data_name)

        # Manage the attributes of the dataset
        variable.attrs.setdefault("units", None)

        variable.attrs.update(dataset_info)
        variable.attrs.update(self._get_global_attributes())
        variable = self._standardize_dims(variable)
        return variable

    @staticmethod
    def _perform_interpolation(variable) -> xr.DataArray:
        """Perform the interpolation from tie points to pixel points.

        Args:
            variable: xarray DataArray containing the dataset to interpolate.

        Returns:
            array containing the interpolate values, all the original metadata
            and the updated dimension names.

        """
        interpolated_values = tie_points_interpolation(
            [variable],
            SCAN_ALT_TIE_POINTS,
            TIE_POINTS_FACTOR
        )[0]
        new_variable = interpolated_values.rename(
            num_tie_points_act="num_pixels",
            num_tie_points_alt="num_lines"
        )
        new_variable.name = variable.name
        new_variable.attrs = variable.attrs
        return new_variable

    @staticmethod
    def _perform_geo_interpolation(longitude, latitude):
        """Perform the interpolation of geographic coodinates from tie points to pixel points.

        Args:
            longitude: xarray DataArray containing the longitude dataset to interpolate.
            latitude: xarray DataArray containing the longitude dataset to interpolate.

        Returns:
            tuple of arrays containing the interpolate values, all the original metadata
                    and the updated dimension names.

        """
        interpolated_longitude, interpolated_latitude = tie_points_geo_interpolation(
            longitude,
            latitude,
            SCAN_ALT_TIE_POINTS,
            TIE_POINTS_FACTOR
        )
        new_longitude = interpolated_longitude.rename(
            num_tie_points_act="num_pixels",
            num_tie_points_alt="num_lines"
        )
        new_longitude.name = longitude.name
        new_longitude.attrs = longitude.attrs
        new_latitude = interpolated_latitude.rename(
            num_tie_points_act="num_pixels",
            num_tie_points_alt="num_lines"
        )
        new_latitude.name = latitude.name
        new_latitude.attrs = latitude.attrs
        return new_longitude, new_latitude

    def _perform_orthorectification(self, variable, orthorect_data_name):
        """Perform the orthorectification."""
        raise NotImplementedError

    def _perform_calibration(self, variable, dataset_info):
        """Perform the calibration."""
        raise NotImplementedError

    def _get_global_attributes(self):
        """Create a dictionary of global attributes to be added to all datasets."""
        attributes = {
            "filename": self.filename,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "spacecraft_name": self.spacecraft_name,
            "ssp_lon": self.ssp_lon,
            "sensor": self.sensor,
            "filename_start_time": self.filename_info["sensing_start_time"],
            "filename_end_time": self.filename_info["sensing_end_time"],
            "platform_name": self.spacecraft_name,
        }

        # Add a "quality_group" item to the dictionary with all the variables and attributes
        # which are found in the 'quality' group of the VII product
        quality_group = self["quality"]
        quality_dict = {}
        for key in quality_group:
            # Add the values (as Numpy array) of each variable in the group where possible
            try:
                quality_dict[key] = quality_group[key].values
            except ValueError:
                quality_dict[key] = None
        # Add the attributes of the quality group
        quality_dict.update(quality_group.attrs)

        attributes["quality_group"] = quality_dict

        return attributes

    def _parse_datetime(self, datetime_str):
        """Parse datetime string using multiple format attempts."""
        for fmt in self.DATETIME_FORMATS:
            try:
                return dt.datetime.strptime(datetime_str, fmt)
            except ValueError:
                continue
        raise ValueError(f"Unrecognized datetime format: {datetime_str}")

    @property
    def start_time(self):
        """Get observation start time."""
        return self._parse_datetime(self["/attr/sensing_start_time_utc"])

    @property
    def end_time(self):
        """Get observation end time."""
        return self._parse_datetime(self["/attr/sensing_end_time_utc"])

    @property
    def spacecraft_name(self):
        """Return spacecraft name."""
        return self["/attr/spacecraft"]

    @property
    def sensor(self):
        """Return sensor."""
        return self["/attr/instrument"]

    @property
    def ssp_lon(self):
        """Return subsatellite point longitude."""
        # This parameter is not applicable to VII
        return None
